/* "boundary.c" */
/* Time-stamp: <97/10/01 11:10:00 hgs> */
/* compute the stopping boundary and prob(stop) */


#include <stdio.h> 
#include <stdlib.h> 
#include <stddef.h>
#include <math.h> 
#include <string.h>
#include <assert.h>

#include "header.h"
#include "cdflib.h"
#define ABSERROR 1.0e-6   /* absolute error of integrator "quanc8" */
#define RELERROR 1.0e-6


/* pass the following arguments of cutoff2() as global variables */
int X, N, X1, X2, N1, N2;
double AS, BS, AE, BE, Delta, AE1, AE2, BE1, BE2, Delta1; 
/* for mixture prior only */
int MIX_K;
double MIX_Design_aS[10][50], WEIGHT[10];  
double MIX_Event_aS[10], MIX_Event_bS[10];

double cutoff3(double p);
double cutoff2(double p);
double cutoff(double p, const int x, const int n, double aS, double bS,
	      double aE, double bE, double delta); 

void compute_design_boundary(Design *design)
{
    Event *comp, *cond;
    for (comp = design->event0->next; comp != NULL; comp = comp->next){
	compute_event_upperbound(comp);	
	compute_event_lowerbound(comp);
	for (cond = comp->child; cond != NULL; cond = cond->next)
	  {  compute_event_upperbound(cond);
	     compute_event_lowerbound(cond);	}
    }
}

void compute_event_upperbound(Event *event)
{
    /* upper bound = Min(X) s.t.  p(etaS + d < etaE |X) > pU */
    /* compute the upper stopping bound x for n patients  */
    /* store the result in event->upperbound[n]  */
    /* starting from n = 5 */
    int i, n, previous, nmin; double aS,bS,aE,bE;
    aS = event->aS;  aE = event->aE;
    bS = event->bS;  bE = event->bE;
    
    /* initialize */
    for (i=0; i<= event->Nmax; i++) event->upperbound[i] = i + 99;
    if (event->boundary_type != 3 && event->boundary_type != 2)
       return;  /* use default boundary */
    
    if ( event->design->Sprior_type > 1 ) { /* mixture prior */
	n = 1;
	while( Lambda_mix(n, n, event->mix_aS, event->mix_bS, 
			  event->design->Sprior_weight, aE, bE, 
			  event->deltaU) < event->pU )
	  ++n;
	event->upperbound[n] = n;
	
	for (i = n+1; i <= event->Nmax; i++) {
	    previous = event->upperbound[i-1];
	    if (Lambda_mix(previous, i, event->mix_aS, event->mix_bS, 
			  event->design->Sprior_weight, aE, bE, 
			  event->deltaU) >= event->pU)
	      event->upperbound[i] = previous;
	    else 
	      event->upperbound[i] = previous + 1;
	}
	nmin = event->design->Nmin;
	if (nmin > 1 && event->parent == NULL) {  
	    /* adjustment for Nmin > 1, but not for conditional event */
	    for (i = nmin - 1; i >= 1; i--) 
	       event->upperbound[i] = event->upperbound[nmin];	    
	}	
    }
    else { /* single prior */
	n = 1;
	while( Lambda4(n, n, aS,bS,aE,bE, event->deltaU) < event->pU )
	  ++n;
	event->upperbound[n] = n;
	
	for (i=n+1; i<= event->Nmax; i++) {
	    previous= event->upperbound[i-1];
	    if (Lambda4(previous, i, aS,bS,aE,bE, event->deltaU) >= event->pU)
	      event->upperbound[i] = previous;
	    else 
	      event->upperbound[i] = previous + 1;
	}
	nmin = event->design->Nmin;
	if (nmin > 1 && event->parent == NULL) {  
	    /* adjustment for Nmin > 1, but not for conditional event */
	    for (i = nmin - 1; i >= 1; i--) 
	      event->upperbound[i] = event->upperbound[nmin];	    
	}
    }
}/*--------------- End of compute_event_upperbound() ----------------------*/


void compute_event_lowerbound(Event *event)
{  
    /* lower bound = Max(X) s.t. p(etaS + d < etaE |X) < pL */
    /* return the lower stopping bound x for n patients     */
    /* store the result in event->lowerbound[n]= x[n]       */
    /* starting from n = 5 */
    int i, n, previous, mm, nmin;  double aS,bS,aE,bE;
    aS = event->aS;  aE = event->aE;
    bS = event->bS;  bE = event->bE;
    
    /* initialize */
     for (i=0; i<= event->Nmax; i++)   event->lowerbound[i] = -99;    
     if (event->boundary_type != 1 && event->boundary_type != 2)
       return; /* use default boundary */

     if ( event->design->Sprior_type > 1 ) { /* mixture prior */
	 n = 1;
	 while ( Lambda_mix(0, n, event->mix_aS, event->mix_bS,
			    event->design->Sprior_weight, aE, bE, 
			    event->deltaL) >= event->pL ) 
	   ++n;	
	 event->lowerbound[n] = 0;  /* first 0 kicks in */

	 for (i= n+1; i<= event->Nmax; i++) {
	     previous = event->lowerbound[i-1];
	     if ( Lambda_mix(previous + 1, i, event->mix_aS, event->mix_bS,
			 event->design->Sprior_weight, aE, bE, 
			 event->deltaL) < event->pL )
	       event->lowerbound[i] = previous + 1;
	     else
	       event->lowerbound[i] = previous;
	 }
	 nmin = event->design->Nmin;
	 if (nmin > 1 /* && event->parent == NULL */) {
	     /* adjustment for Nmin > 1 --> "run-back" rules */
	     mm = event->lowerbound[nmin];
	     for (i = nmin - 1; i >= 1; i--) {
		 event->lowerbound[i] = mm-1;
		 --mm;
	     }
	 }	
     }
     else { /* single prior */
	 n = 1;
	 while ( Lambda4(0, n, aS,bS,aE,bE, event->deltaL) >= event->pL ) 
	   ++n;	
	 event->lowerbound[n] = 0;  /* first 0 kicks in */

	 for (i= n+1; i<= event->Nmax; i++) {
	     previous = event->lowerbound[i-1];
	     if ( Lambda4(previous + 1, i, aS,bS,aE,bE, event->deltaL) 
		< event->pL )
	       event->lowerbound[i] = previous + 1;
	     else
	       event->lowerbound[i] = previous;
	 }
	 nmin = event->design->Nmin;
	 if (nmin > 1 /*&& event->parent == NULL*/) {
	     /* adjustment for Nmin > 1 --> "run-back" rules */
	     mm = event->lowerbound[nmin];
	     for (i = nmin - 1; i >= 1; i--) {
		 event->lowerbound[i] = mm-1;
		 --mm;
	     }
	 }
     }
 }/*-------------- End of compute_event_lowerbound() ---------------------*/

 /* return 1 if success; 0 if fail */
void input_boundary_parameter(Event *event)
{
    int j, option;
    if (event->boundary_type != 4) {
	printf("\n-----------------------------------------------\n"
	       "Current boundary parameters of event: '%s'\n", event->name);
	j = event->boundary_type;
	printf("boundary type: %s\n", (j==1)? "lower efficacy" : 
	       (j==2)? "upper and lower" : 
	       (j==3)? "upper adverse" : "no monitoring boundary");
	if (event->boundary_type == 1 && event->boundary_type == 2)
	  printf("delta_L = %.3f,  p_L = %.3f\n", event->deltaL, event->pL);
	if (event->boundary_type == 2 && event->boundary_type == 3)
	  printf("delta_U = %.3f,  p_U = %.3f\n", event->deltaU, event->pU);
	printf("-----------------------------------------------\n\n");
    }
    printf("---------------------------------------------------------\n"
	   "Please select the type of boundary for event: '%s'.\n",
	   event->name);  
    printf("---------------------------------------------------------\n\n"
	   "(1) Lower efficacy only.\n"
	   "    The decision criterion is to stop if\n"
	   "    Pr[Theta_S + delta < Theta_E | Data ] < p_L, delta >= 0\n\n"
	   "(2) Upper adverse only.\n"
	   "    The decision criterion is to stop if\n"
	   "    Pr[Theta_S + delta < Theta_E | Data ] > p_U\n"
	   "    A positive value of delta is the maximum acceptable\n"
	   "    increase (slippage) for the rate of this adverse event.\n"
	   "    A negative value of delta is the magnitude of targeted\n"
	   "    improvement achieved by E over S for this adverse event.\n\n"
	   "(3) Upper & lower bounds.\n"
	   "    Apply both upper and lower stopping criteria for this event.\n"
	   "    Please use delta_L >= 0 and delta_U >= 0 in this case.\n\n"
	   "(4) No monitoring boundary for this event.\n\n"
	   "(5) Skip this menu.\n");
    do {
	printf("\n(Please choose 1 to 5)\n> ");
	option = get_an_integer();
    } while (option < 1 || option > 5);
    if ( option == 1 ) {
	printf("\ndelta_L: > ");
	event->deltaL = get_a_double(0e0, 1e0);
	if ( (int)event->deltaL == -2 )	{event->boundary_type = 4; return;}
	printf("\np_L: > ");
	event->pL = get_a_double(0e0, 1e0);
	if ( (int)event->pL == -2 )	{event->boundary_type = 4; return;}
	event->boundary_type = 1; event->type = 1;
    }
    else if ( option == 2 ) { /* upper only */
	printf("\ndelta_U: > ");
	event->deltaU = get_a_double(-1e0, 1e0); /* allow delta < 0 */
	if ( (int)event->deltaU == -2 )	{event->boundary_type = 4; return;}
	printf("\np_U: > ");
	event->pU =  get_a_double(0e0, 1e0);
	if ( (int)event->pU == -2 )	{event->boundary_type = 4; return;}
	event->boundary_type = 3; event->type = 0;
    }
    else if ( option == 3 ) { /* both */
	printf("\ndelta_L: > ");
	event->deltaL = get_a_double(0e0, 1e0);
	if ( (int)event->deltaL == -2 )	{event->boundary_type = 4; return;}
	printf("\np_L: > ");
	event->pL = get_a_double(0e0, 1e0);
	if ( (int)event->pL == -2 )	{event->boundary_type = 4; return;}
	printf("\ndelta_U: > ");
	event->deltaU = get_a_double(0e0, 1e0);
	if ( (int)event->deltaU == -2 )	{event->boundary_type = 4; return;}
	printf("\np_U: > ");
	event->pU =  get_a_double(0e0, 1e0);	
	if ( (int)event->pU == -2 )	{event->boundary_type = 4; return;}
	event->boundary_type = 2;   event->type = 1;
    }
    else if (option == 4) {	
	event->deltaL = 0e0; event->pL = 0e0;
	event->deltaU = 0e0; event->pU = 0e0;
	event->boundary_type = 4;  
    }
    printf("\n");
}/*------------------------------end of input_boundary_paremeter()----*/


 /* print out the boundary parameters on stdout (screen) */
void review_boundary_parameter(Design *design)
{
    Event *comp, *cond; 
    char buf1[30], buf2[30], buf3[30], buf4[30];
    int j, count;
    strcpy(buf1, "Lower Efficacy ");
    strcpy(buf2, "Upper and Lower");
    strcpy(buf3, "Upper Adverse  "); 
    strcpy(buf4, "No  Boundary   ");
    
    printf("\n-----------------------------\n"
	   "Current Boundary parameters\n"
	   "-----------------------------\n"
	   " Name           Boundary type     deltaL   pL     deltaU   pU\n"
	   "---------      ---------------   -------------   -------------\n");
    count = 0;
    for (comp = design->event0->next; comp != NULL; comp = comp->next) 
      {
	  j = comp->boundary_type;
	  printf("%d) %-10s\t", ++count, comp->name);
	  if (j == 1 )
	    printf("%s  %.3lf  %.3lf    --      -- ", buf1,
		   comp->deltaL, comp->pL);
	  if (j == 2 )
	    printf("%s  %.3lf  %.3lf   %.3lf  %.3lf", buf2,
		   comp->deltaL, comp->pL, comp->deltaU, comp->pU);
	  if (j == 3)
	    printf("%s  --     --      %.3lf  %.3lf", buf3,
		   comp->deltaU, comp->pU);
	  if (j == 4)
	    printf("%s  --     --       --      -- ", buf4);
	  printf("\n\n");  
      }  
    assert(count == design->number_of_event);
    /* conditional events */
    for (comp = design->event0->next; comp != NULL; comp = comp->next) 
      for (cond = comp->child; cond != NULL; cond = cond->next) { 
	  j = cond->boundary_type;
	  printf("\n%d) %-10s\t", ++count, cond->name);
	  if (j == 1 )
	    printf("%s  %.3lf  %.3lf    --      -- ", buf1,
		   cond->deltaL, cond->pL);
	  if (j == 2 )
	    printf("%s  %.3lf  %.3lf   %.3lf  %.3lf", buf2,
		   cond->deltaL, cond->pL, cond->deltaU, cond->pU);
	  if (j == 3)
	    printf("%s  --     --      %.3lf  %.3lf", buf3,
		   cond->deltaU, cond->pU);
	  if (j == 4)
	    printf("%s  --     --       --      -- ", buf4);
	  printf("\n\n");  
      }  
    assert(count == design->number_of_event + 
	   design->number_of_cond_event);
}/*----------------------------end of review_boundary_parameter()------*/


/* For Phase II A  only */
double Lambda3(int x, int n, double aE, double bE, double p_Null)
{
    /* return Pr[ theta > pNull | X/n ] */
    return 1.0 - (double) pbeta(p_Null, aE + x, bE + n - x);   
}


/* using new integrator new_quanc8() */
double Lambda4(int x, int n, double aS, double bS, double aE, double bE,
	       double delta)
{    
    int nofun; double estimate_error = .0, flag = .0;
    double result = .0;
    X = x; N = n;
    AS = aS; BS = bS; AE = aE; BE = bE; Delta = delta;
    new_quanc8((*cutoff2), 0e0, 1e0-delta, ABSERROR, RELERROR, &result,
	       &estimate_error, &nofun, &flag);
    if ( estimate_error > ABSERROR ) 
      printf("** Warning: Estimated error = %.3e\n", estimate_error); 
    return Min(1e0, result);
/*--------------------------end of Lambda4() ------------*/
}


/* for comparing two randomized arms -> an extension of randomized trial */
double Lambda5(int x1, int x2, int n1, int n2, double ae1, double be1, 
	       double ae2, double be2, double delta)
{
/*-------------------------------------------------------------
  pr( theta1 + delta < theta2 | data )
  where both theta1 and theta2 are updated sequentally
--------------------------------------------------------------*/
    int nofun; double estimate_error = .0, flag = .0;
    double result = .0;
    N1 = n1; N2 = n2; 
    X1 = x1; X2 = x2;
    AE1 = ae1; BE1 = be1; 
    AE2 = ae2; BE2 = be2; 
    Delta1 = delta;

    new_quanc8((*cutoff3), 0e0, 1e0-delta, ABSERROR, RELERROR, &result,
	       &estimate_error, &nofun, &flag);
    if ( estimate_error > ABSERROR ) 
      printf("** Warning: Estimated error = %.3e\n", estimate_error); 
    return Min(1e0, result);
}


/* for the mixture prior of thetaS */
double Lambda_mix(int x, int n, double *aS, double *bS, double *weight,
		  double aE, double bE, double delta)
{
    int i; double result = 0e0;
    for (i = 0; i< MIX_K; i++) /* MIX_K = number of arms */
      result += weight[i] * Lambda4(x,n,aS[i],bS[i],aE,bE,delta);
    return result;
}/*-------------------end of Lambda_mix() ------------*/


/* integrand */
/* pass the rest of the arguments as global variables */
double cutoff2(double p)
{  
    double integrand;
    /* boundary checking is handled by pbeta() and dbeta() */
    integrand = ((1e0 - pbeta(p + Delta, AE + X, BE + N - X))* 
		 dbeta(p, AS, BS) );
     return integrand;
/*------------------------ End of cutoff2() ----------------------------*/
} 

/* integrand for the comparison between two arms in the randomized trial */
/* pass the rest of the arguments as global variables from the 
   'parent' function */
double cutoff3( double p )
{  
    double integrand;
    /* boundary checking is handled by pbeta() and dbeta() */
    integrand = ((1.0 - pbeta(p + Delta1, AE2 + X2, BE2 + N2 - X2) ) * 
		 dbeta(p, BE1 + X1, BE1 + N1 - X1) );

     return integrand;
}/*------------------------ End of cutoff2() ----------------------------*/


/* integrand --> pass all arguments as local variables */
double cutoff(double p, const int x, const int n, 
	      double aS, double bS, double aE, double bE, double delta) 
{  
    double integrand;
    /* double aS = event->aS; double bS = event->bS; double aE =
       event->aE; double bE = event->bE; */
    /* if (p+delta > 1e0) p = 1e0 - delta;     */ 
    integrand = ((1.0 - pbeta(p + delta, aE + x, bE + n - x))* 
		 dbeta(p, aS, bS) );
    return integrand;
}


/* draw the plot of func(d) vs d  for d = a to b */
void plot_func(double a, double b, int x, int n, 
	       Event *event, double (*func)(), char *filename)
{
    double dd;  double temp;  FILE *fp;
    if ( a > b ) { temp = a;  a = b; b = temp; }  /* swap(a, b) */
    if ((fp = fopen(filename, "w")) == NULL) {
	printf("cannot open file %s in plot_func()\n", filename);
	exit(0);
    }
    printf("\n** Data (delta-lambda) stored in file: \"%s\" **\n\n", filename);
    fprintf(fp, "delta    lambda\n");
    dd = a;
    while(dd <= b + .001) {
	fprintf(fp, "%.4lf  %lf\n", dd, func(x, n, event, dd));
	/* dd += (double) (b-a)/100; */
	dd += .005;
    }
    (void)fclose(fp);
}/*--------------End of plot_func() -------------------------------*/


void read_design_boundary(Design *design)
{
    Event *comp, *cond; int option;
    for (comp=design->event0->next; comp!=NULL; comp=comp->next)
      read_event_boundary(comp);
    if (design->number_of_cond_event != 0) {
	  for (comp=design->event0->next; comp!=NULL; comp=comp->next)
	    for (cond=comp->child; cond!=NULL; cond=cond->next)
	      read_event_boundary(cond);
      }
    printf("\nConfirmed?(y/n) > ");
    option = get_an_integer();
    if (option == YES)   print_boundary_in_file(design);
    else read_design_boundary(design);
}/*-----------------------end of read_design_boundary()------------*/


/* Maybe better to input only the check points, do it later */
void read_event_boundary(Event *event)
{
    int i, j, k, bnd;
    printf("read boundary for event: %s\n", event->name); 
    /* i = 0;
    while ( bound < event->Nmax ) {
	printf("Stop at %d out of: > ", i);
	bound = get_an_integer();
	event->lowerbound[bound] = i;    }*/

    for (i = event->design->Nmin; i< event->Nmax; i++) {
	printf("boundary at total number of patients %d:\n", i); 
	if (event->boundary_type == 1 || event->boundary_type == 2) { 
	    k = (i == event->design->Nmin)? 0 : event->lowerbound[i-1];
	    printf("lower bound (>= %d, <= %d): ", k, i);
	    bnd = get_an_integer();
	    if (bnd >= k && bnd <= i)  event->lowerbound[i] = bnd;
	    else {
	      --i; continue;
	    }
	}
	if (event->boundary_type == 3 || event->boundary_type == 2) { 
	    j = (i==event->design->Nmin)? 0 : event->upperbound[i-1];
	    printf("upper bound (>= %d, <= %d) ", j, i);
	    bnd = get_an_integer();
	    if (bnd >= j && bnd <= i) event->upperbound[i] = bnd;
	    else {
	      --i; continue;
	    }
	}
    }
}/*--------------- End of read_event_boundary() ----------------------*/

