/* "compute.c" */
/* Time-stamp: <97/08/13 18:48:32 hgs> */
/* numerical integration routine and some utility functions */
  
#include <stdio.h> 
#include <stdlib.h> 
#include <stddef.h>
#include <math.h> 
#include <string.h>
#include <assert.h>
#include <time.h>
#include <float.h>

#include "header.h"
#include "ranlib.h"
#include "cdflib.h"
#include "zrorj.h" /* for stzror(), and zror() used in qbeta() */

#define Zero DBL_EPSILON


/* NOTE: integrator using 8-panel Newton-Cotes rule. obtain fortran
   codes from Rene Gonin, translated to C by translator "f2c".
   fun= cutoff(p,X,n,aS,bS,aE,bE,delta) integrated over p = *a to p = *b */
int quanc8(double (*fun)(), int X, int n,
	   double aS, double bS, double aE, double bE,
	   double delta,
	   double *a, double *b, double *abserr, double *relerr, 
	   double *result, double *errest, int *nofun, double *flag)
{
    /* System generated locals */
    double d_1, d_2;
    /* Local variables */
    static double area, cor11, temp, step, qnow, f[16];
    static int i, j;
    static double x[16], qdiff, fsave[240];	/* 1-dim of [8][30] */
    static int nofin;
    static double qleft;
    static double xsave[240], f0, stone, qprev, w0, w1, w2, w3, w4, x0;
    static int levmin, levmax;
    static double qright[31], esterr, tolerr;
    static int levout, nim, lev;
/*
    estimate the integral of fun(x) from a to b 
    to a user provided tolerance.               
    an automatic adaptive routine based on      
    the 8-panel newton-cotes rule.             

    input..                                                    
    fun the name of the integrand function subprogram fun(x).   
    a the lower limit of integration.                           
    b the upper limit of integration.(b may be less than a.)    
    relerr a relative error tolerance.(should be non-negative)  
    abserr an absolute error tolerance.(should be non-negative) 
                                                                
    output..                                                    
    nofun the number of function values used in calculation of result. 
    flag a reliability indicator.if flag is zero,then result    
    probably satisfies the error tolerance.if flag is           
    xxx.yyy,then xxx=the number of intervals which have         
    not converged and 0.yyy=the fraction of the interval        
    left to do when the limit on nofun was approached.          
                                                                
    *** stage 1 *** general initialization           */          
/*    set constants.      */
    levmin = 1;
    levmax = 30;
    levout = 6;
/*    nomax = 5000; */
/*    i_1 = levout + 1; */
/*    nofin = nomax - 8 * (levmax - levout + pow(2, levout+1)); */
/*    nofin = 5000 - 8 * ( 30 - 6 + pow(2, 6+1)); */
    nofin = 3784;
/*                                                                  0680*/
/*    trouble when nofun reaches nofin                              0690*/
/*                                                                  0700*/
    w0 = .27908289241622575;
    w1 = 1.6615167548500882;
    w2 = -.26186948853615521;
    w3 = 2.9618342151675483;
    w4 = -1.2811287477954145;
/*    initialize running sums to zero.                              0760*/
    *flag = 0.;
    *result = 0.;
    cor11 = 0.;
    *errest = 0.;
    area = 0.;
    *nofun = 0;
    if (*a == *b) {	return 0;    }
/*    *** stage 2 *** initialization for first interval             0850*/
    lev = 0;
    nim = 1;
    x0 = *a;
    x[15] = *b;
    qprev = 0.;
    f0 = (*fun)(x0, X,n,aS,bS,aE,bE,delta);
    stone = (*b - *a) / 16.;
    x[7] = (x0 + x[15]) / 2.;
    x[3] = (x0 + x[7]) / 2.;
    x[11] = (x[7] + x[15]) / 2.;
    x[1] = (x0 + x[3]) / 2.;
    x[5] = (x[3] + x[7]) / 2.;
    x[9] = (x[7] + x[11]) / 2.;
    x[13] = (x[11] + x[15]) / 2.;
    for (j = 2; j <= 16; j += 2) {
	f[j - 1] = (*fun)(x[j - 1], X,n,aS,bS,aE,bE,delta);
/* L25: */
    }
    *nofun = 9;
/*                                                                     1050*/
/*    *** stage 3 *** central calculation                              1060*/
/*    requires qprev,x0,x2,x4,...,x16,f0,f2,f4,...,f16.                1070*/
/*    calculates x1,x3,....x15, f1,f3,...f15.qleft,qright,qnow,qdiff,  1080*/
/*    qdiff,area                                                       1090*/
L30:
    x[0] = (x0 + x[1]) / 2.;
    f[0] = (*fun)(x[0],  X,n,aS,bS,aE,bE,delta);
    for (j = 3; j <= 15; j += 2) {
	x[j - 1] = (x[j - 2] + x[j]) / 2.;
	f[j - 1] = (*fun)(x[j - 1], X,n,aS,bS,aE,bE,delta);
/* L35: */
    }
    *nofun += 8;
    step = (x[15] - x0) / 16.;
    qleft = (w0 * (f0 + f[7]) + w1 * (f[0] + f[6]) + w2 * (f[1] + f[5]) + w3 *
	     (f[2] + f[4]) + w4 * f[3]) * step;
    qright[lev] = (w0 * (f[7] + f[15]) + w1 * (f[8] + f[14]) + w2 * (f[9] + f[
	    13]) + w3 * (f[10] + f[12]) + w4 * f[11]) * step;
    qnow = qleft + qright[lev];
    qdiff = qnow - qprev;
    area += qdiff;
/*                                                                  1250*/
/*    *** stage 4  *** interval convergence test                    1260*/
/*                                                                  1270*/
    esterr = fabs(qdiff) / 1023.;
/* Computing MAX */
    d_1 = *abserr, d_2 = *relerr * fabs(area);
    tolerr = Max(d_1,d_2) * (step / stone);
    if (lev < levmin) {
	goto L50;
    }
    if (lev >= levmax) {
	goto L62;
    }
    if (*nofun > nofin) {
	goto L60;
    }
    if (esterr <= tolerr) {
	goto L70;
    }
/*                                                                  1340*/
/*    *** stage 5 *** no convergence                                1350*/
/*    locate next interval.                                         1360*/
/*                                                                  1370*/
L50:
    nim <<= 1;
    ++lev;
/*                                                                  1400*/
/*    store right hand elements for future use.                     1410*/
/*                                                                  1420*/
    for (i = 1; i <= 8; ++i) {
	fsave[i + (lev << 3) - 9] = f[i + 7];
	xsave[i + (lev << 3) - 9] = x[i + 7];
/* L52: */
    }
/*                                                                  1470*/
/*    assemble left hand elements for immediate use.                1480*/
/*                                                                  1490*/
    qprev = qleft;
    for (i = 1; i <= 8; ++i) {
	j = -i;
	f[(j << 1) + 17] = f[j + 8];
	x[(j << 1) + 17] = x[j + 8];
/* L55: */
    }
    goto L30;
/*                                                                  1570*/
/*    *** stage 6 *** trouble section                               1580*/
/*    number of function values is about to exceed limit.           1590*/
/*                                                                  1600*/
L60:
    nofin <<= 1;
    levmax = levout;
    *flag += (*b - x0) / (*b - *a);
    goto L70;
/*                                                                  1650*/
/*    current level is levmax.                                      1660*/
/*                                                                  1670*/
L62:
    *flag += 1.;
/*                                                                  1690*/
/*    *** stage 7 *** interval converged                            1700*/
/*    add contributions into running sums.                          1710*/
/*                                                                  1720*/
L70:
    *result += qnow;
    *errest += esterr;
    cor11 += qdiff / 1023.;
/*                                                                  1760*/
/*    locate next interval.                                         1770*/
/*                                                                  1780*/
L72:
    if (nim == nim / 2 << 1) {
	goto L75;
    }
    nim /= 2;
    --lev;
    goto L72;
L75:
    ++nim;
    if (lev <= 0) {
	goto L80;
    }
/*                                                                  1850*/
/*    assemble elements equired for the next interval.              1860*/
/*                                                                  1870*/
    qprev = qright[lev - 1];
    x0 = x[15];
    f0 = f[15];
    for (i = 1; i <= 8; ++i) {
	f[(i << 1) - 1] = fsave[i + (lev << 3) - 9];
	x[(i << 1) - 1] = xsave[i + (lev << 3) - 9];
/* L78: */
    }
    goto L30;
/*                                                                  1960*/
/*    ...  stage 8  ***  finalize and return                        1970*/
/*                                                                  1980*/
L80:
    *result += cor11;
/*                                                                  2000*/
/*    male sure errest not less then roundoff level.                2010*/
/*                                                                  2020*/
    if (*errest == 0.) {
	return 0;
    }
L82:
    temp = fabs(*result) + *errest;
    if (temp != fabs(*result)) {
	return 0;
    }
    *errest *= 2.;
    goto L82;
}/*---------------------------------end of quanc8()--------------------*/

/* modified version, should be better than quanc8(), a bit more
   elegant. currently used by Lambda4() */
void new_quanc8(double (*fun)(), double a, double b, 
	       double abserr, double relerr, double *result,
	       double *errest, int *nofun, double *flag)
{
    double qright[31],f[16],x[16],fsave[30][8],xsave[30][8];
    int levelmin,levelmax,levelout,nofin, i, j,level, nim;
    double w0,w1,w2,w3,w4,cor11,area,  x0,qprev,f0;
    double step,qleft,qnow,qdiff,esterr,tolerr;
/*
     estimate the integral of fun(x) from a to b                       360
     to a user provided tolerance.                                     370
     an automatic adaptive routine based on                            380
     the 8-panel newton-cotes rule.                                    390
                                                                       400
     input..                                                           410
     fun the name of the integrand function subprogram fun(x).         420
     a the lower limit of integration.                                 430
     b the upper limit of integration.(b may be less than a.)          440
     relerr a relative error tolerance.(should be non-negative)        450
     abserr an absolute error tolerance.(should be non-negative)       460
                                                                       470
     output..                                                          480
                                                                       490
     result an approximation to the integral hopefully satisfying the  500
     least stringent of the two error tolerances.                      510
     errest an estimate of the magnitude of the actual error.          520
     nofun the number of function values used in calculation of result.530
     flag a reliability indicator.if flag is zero,then result          540
     probably satisfies the error tolerance.if flag is                 550
     xxx.yyy,then xxx=the number of intervals which have               560
     not converged and 0.yyy=the fraction of the interval              570
     left to do when the limit on nofun was approached.                580
                                                                       590
     *** stage 1 *** general initialization                            600
     set constants.                                                    610
                                                                       620*/
    levelmin = 1;    levelmax = 30;     levelout = 6;
    /* nomax = 5000;
       nofin = nomax - 8 * (levelmax - levelout + pow(2,levelout + 1)); */
    nofin = 3784; 
    /*
     trouble when nofun reaches nofin        */
    w0 = .27908289241622575;
    w1 = 1.6615167548500882;
    w2 = -.26186948853615521;
    w3 = 2.9618342151675483;
    w4 = -1.2811287477954145;
/*
       initialize running sums to zero.                                  760
*/
    *flag = cor11 = *errest = area = 0e0;
    *result = 0e0;
    *nofun = 0;
    if (a == b) return; 
/*
     *** stage 2 *** initialization for first interval                 850
*/
    level = 0;     nim = 1;     
    x0 = a;    x[15] = b;     qprev = 0e0;

    f0 = (*fun)(x0); 
/*    stone = (b - a) / 16e0; */

    x[7] = (x0 + x[15]) / 2e0;
    x[3] = (x0 + x[7]) / 2e0;
    x[11] = (x[7] + x[15]) / 2e0;
    x[1] = (x0 + x[3]) / 2e0;
    x[5] = (x[3] + x[7]) / 2e0;
    x[9] = (x[7] + x[11]) / 2e0;
    x[13] = (x[11] + x[15]) / 2e0;
    for (j=2; j<=16; j+=2) f[j - 1] = (*fun)(x[j - 1]);

    *nofun = 9;
S30:
/*     *** stage 3 *** central calculation                             
     requires qprev,x0,x2,x4,...,x16,f0,f2,f4,...,f16.                 
     calculates x1,x3,....x15, f1,f3,...f15.qleft,qright,qnow,qdiff,area */
    x[0] = (x0 + x[1]) / 2e0;
    f[0] = (*fun)(x[0]);
  
    for (j=3; j<=15; j+=2) {
        x[j - 1] = (x[j - 1 - 1] + x[j + 1 - 1]) / 2e0;
   	f[j - 1] = (*fun)(x[j - 1]);
    }

    *nofun += 8;
    step = (x[15] - x0) / 16e0;
    qleft = (w0*(f0 + f[7]) + w1*(f[0] + f[6]) + w2*(f[1] + f[5]) + 
	     w3*(f[2] + f[4]) + w4*f[3])*step;
    qright[level] = (w0*(f[7] + f[15]) + w1*(f[8] + f[14]) + 
		   w2*(f[9] + f[13]) + w3*(f[10] + f[12]) + w4*f[11]) * step;
    qnow = qleft + qright[level];
    qdiff = qnow - qprev;
    area += qdiff;
/*
     *** stage 4  *** interval convergence test   
*/
    esterr = fabs(qdiff) / 1023e0; 
    /* tolerr =  Max(abserr, relerr * fabs(area)) * (step / stone); */
    /* (step / stone) = (x[15] - x0)/(b-a); */
    tolerr =  Max(abserr, relerr * fabs(area)) * (x[15] - x0)/(b-a);
  
    if (level < levelmin) goto S50;
    if (level >= levelmax) goto S62;
    if (*nofun > nofin) goto S60;
    if (esterr <= tolerr) goto S70;
S50:
    /*   *** stage 5 *** no convergence 
	 locate next interval.   
	 */
    nim = 2*nim;     level += 1;
/*
     store right hand elements for future use.                         1410
 */
    for (i=1; i<=8; i++) {
        fsave[level - 1][i - 1] = f[i + 7];
        xsave[level - 1][i - 1] = x[i + 7];
    }
/*
                                                                        1470
     assemble left hand elements for immediate use.                    1480
                                                                       1490
*/
    qprev = qleft;
    for (i=1; i<=8; i++) {
	/*        j = -i;   f[2*j + 18 - 1] = f[j + 9 - 1];
		  x[2*j + 18 - 1] = x[j + 9 - 1]; */
	f[17-2*i] = f[8-i];
	x[17-2*i] = x[8-i];
    }
    goto S30;
S60:
/*
                                                                       1570
     *** stage 6 *** trouble section                                   1580
     number of function values is about to exceed limit.               1590
                                                                       1600
*/
    nofin = 2*nofin;
    levelmax = levelout;
    *flag += ((b - x0) / (b - a));
    goto S70;
S62:
/*
     current level is levelmax.      */
    *flag += 1e0;
S70:
/*
                                                                       1690
     *** stage 7 *** interval converged                                1700
     add contributions into running sums.                              1710
                                                                       1720
*/
    *result += qnow;
    *errest += esterr;
    cor11 += (qdiff / 1023e0);
S72:
/*
                                                                       1760
     locate next interval.                                             1770
                                                                       1780
*/
    if (nim == 2*(nim / 2)) goto S75;
    nim /= 2;
    level -= 1;
    goto S72;
S75:
    nim += 1;
    if (level <= 0) goto S80;
/*
                                                                       1850
     assemble elements equired for the next interval.                  1860
                                                                       1870
*/
    qprev = qright[level - 1];
    x0 = x[15];
    f0 = f[15];
    for (i=1; i<=8; i++) {
        f[2*i - 1] = fsave[level - 1][i - 1];
        x[2*i - 1] = xsave[level - 1][i - 1];
    }
    goto S30;
S80:
/*     ...  stage 8  ***  finalize and return  */
    *result += cor11;
/*     make sure errest not less then roundoff level. */
    if (*errest == 0e0)	return; 

    while((fabs(*result) + *errest) == fabs(*result))
      *errest *= 2e0;
    return; 
    /*   temp = fabs(*result) + *errest;
	 if (temp != fabs(*result))	return;
	 *errest = 2e0 * (*errest);
	 goto S82; */
    
}/*------------------------------------end of new_quanc8()--------*/



double qbeta(double p, double a, double b)
{
    int status;
    /* data type is crucial. double and int won't work*/
    float xlo, xhi, abstol, reltol, x, fx;
    unsigned long qleft, qhi;

    status = 0;
    xlo = 0.0;
    xhi = 1.0;
    abstol = .0001;
    reltol = 0.0;
    x = 0.0;
    fx = 0.0;
    qleft = 0;
    qhi = 0;
 
    stzror(&xlo,&xhi,&abstol,&reltol);

    zror(&status,&x,&fx,&xlo,&xhi,&qleft,&qhi);

    while (status==1) {
        zror(&status,&x,&fx,&xlo,&xhi,&qleft,&qhi);
        if (status==1) fx = (float) (pbeta((double) x, a,b)-p);
    }
    return(xlo);
}



/* c.d.f Beta(a, b, z): pr(x < z) s.t.  x ~ beta(a,b)  */ 
double pbeta(double z, double a, double b)
{
    double prob, comp, y;
    y = 1.0 - z;
    cumbet(&z, &y, &a, &b, &prob, &comp); /* in cdflib.c */
    return prob;
}

double dbeta(double z, double a, double b)
{
    double density, temp, p, q;
    if ( z <= 0e0 || z >= 1e0 )  return 0e0;
    temp = betaln(&a, &b);
    p = (a - 1.0) * log(z);
    q = (b - 1.0) * log(1.0 - z);
    density = exp( p + q - temp);
    return density;
}


/* reparameterization */
/* from mean, width, and cover prob to Dirichlet a[i]  */
void getdir(Design *design, double width, double coverp, int flag)
{
    int i;
    double high = 10000.0;
    double low = 0.0000001;
    double mid, sum, hi_eval, lo_eval, mid_eval;
    double mu = 0.0;
    double half = width / 2.0;
    Event *temp = design->current_event;
    
    if (flag == 1) {
	for (i=0; i<design->number_of_element; i++) 
	  if (temp->definition[i]==1) mu += design->muS[i];
    }
    else if (flag == 2) {
	for (i=0; i<design->number_of_element; i++) 
          if (temp->definition[i]==1) mu += design->muE[i];
    }

    hi_eval = pbeta(mu+half, high, high*(1.0-mu)/mu) -
             pbeta(mu-half, high, high*(1.0-mu)/mu) - coverp;
    lo_eval = pbeta(mu+half, low, low*(1.0-mu)/mu) - 
             pbeta(mu-half, low, low*(1.0-mu)/mu) - coverp;
    
    mid_eval = hi_eval;
    
    /* binary search for the right beta(a,b) s.t. a/(a+b) = mu 
       error +/- 1.0e-7 */
    while ((mid_eval< -.00000001)||(mid_eval>.00000001)) {
        mid = (high+low)/2;
        mid_eval = pbeta(mu+half, mid, mid*(1.0-mu)/mu) -
	           pbeta(mu-half, mid, mid*(1.0-mu)/mu) - coverp;
        if ((mid_eval * lo_eval) < Zero) {
            high = mid;
            hi_eval = mid_eval;
        } else if ((mid_eval * hi_eval) < Zero) {
            low = mid;
            lo_eval = mid_eval;
        } else printf("routine getdir() doesn't converge.\n");
    }

    sum = mid / mu; 
    
    if (flag == 2) {
        for (i=0; i<design->number_of_element; i++)  
	  design->aE[i] = design->muE[i] * sum;
    }
    else if (flag == 1) {
        for (i=0; i<design->number_of_element; i++)  
	  design->aS[i] = design->muS[i] * sum;
    }
}/*--------------------------------------end of getdir()----*/


/* compute Nmax */
int compute_Nmax(Event *event, double mu, double width, double coverprob)
{
    double aS,bS,aE,bE, muS, lobound, hibound;
    double up, lo, xE,xl,xc, fl,fc, a,b;
    int n;
    
    aS = event->aS;    bS = event->bS;
    aE = event->aE;    bE = event->bE;
    muS = (mu == 0.0)?  aS/(aS+bS) : mu;
    lobound = (1 - coverprob)/2;
    hibound = (1 + coverprob)/2;

    n = 1; 
    lo = 0.0; up = width + 1.0; /* an initial setup to start the
				   while() loop */
    while ( up - lo > width ) {
        xE = muS *(aE+bE+n)-aE;
        xl = floor(xE);
        xc = ceil(xE);
        fl = (xl+aE)/(n+aE+bE);
        fc = (xc+aE)/(n+aE+bE);
	
        /* if (fabs((double)(fl-.5))<fabs((double)(fc-.5))) xE =xl; */
	/* why abs ? */
	if ( fabs(fl - 0.5) < fabs(fc - 0.5))  xE =xl; 
	else xE=xc;
	
        a = aE + xE;
	b = bE + n - xE;
	
        up = qbeta(hibound,a,b);
        lo = qbeta(lobound,a,b);
	/*       if (up - lo <= width) {   ans = n;   break; } */
	++n;
    }
    
    return (n - 1);   /* design->Nmax = n - 1 */
}/*-----------------------------------end of compute_Nmax()--------*/


/* sort the key[] and rearranging index[] correspondingly */
void qrank(double *key, int *index, int left, int right)
{
    int i, last;
    if ( left >= right ) return;
    dswap(key, left, (left+right)/2);
    iswap(index, left, (left+right)/2);
    last = left;
    for (i = left+1; i<= right; i++) 
      if (key[i] < key[left]) {
	  ++last;
	  dswap(key, last, i); iswap(index, last, i);
      }
    dswap(key, left, last); iswap(index, left, last);
    qrank(key, index, left, last-1);
    qrank(key, index, last+1, right);
}/*---------------------------------------end of qrank()----*/

void dswap(double *arr, int i, int j)
{
    double temp;
    temp = arr[i]; arr[i] = arr[j]; arr[j] = temp;
}

void iswap(int *arr, int i, int j)
{
    int temp;
     temp = arr[i]; arr[i] = arr[j]; arr[j] = temp;
}
