/* "random.trial.c" */     /* Time-stamp: <97/12/19 12:06:01 hgs> */
/* running randomized trial */
  
#include <stdio.h> 
#include <stdlib.h> 
#include <stddef.h>
#include <math.h> 
#include <string.h>
#include <assert.h>
#include <time.h>
#include <float.h>

#include "header.h"
#include "ranlib.h"

#define MERGE   NO /* set up 2:1 randomized scheme */

/* balanced randomization scheme[i] ~ {0,1,2,...,k-1} 
   for k arms, total patient N */
int *balrand(int narms, int N)
{
    int m1, i, k, *scheme; double *sample;
    scheme = (int*) malloc( N * sizeof(int));
    if (!scheme) return scheme;
    sample = (double*) malloc( N * sizeof(double));
    if (!sample) { 
	printf("Cannot create randomization scheme.\n");
	free(scheme);
	return NULL;
    }
    if (MERGE == YES) k = narms + 1;
    else k = narms;

    m1 = (int) ceil((double)N/k/2);  /* m2 = N/k; */
    for (i=0; i< m1 * k; i++) {
	sample[i] = (double) genunf((float)0.0, (float)1.0);
	scheme[i] = (i % k);
    }
    qrank(sample, scheme, 0, m1*k-1);    
    for (i = m1*k; i< N; i++) {
	sample[i] = (double) genunf((float)0.0, (float)1.0);
	scheme[i] = (i % k);
    }
    qrank(sample, scheme, m1*k, N-1);
    free(sample);
    return scheme;
}/*-----------------end of balrand()--------------*/

/* update the assign[] given by balrand() when arm k is terminated.
   hit[i] = NO means arm i is not terminated. n = current #patients.
   return 0 if all arms are terminated */
int re_assign(int k, int total_arms, int n, int N, int *assign, int *hit)
{
    int i, j, arms[10], arms_remain = 0, rest = 0; double *sample;
    int *index;
    j = 0;
    for (i=0; i< total_arms; i++) 
      if (hit[i] == NO) { ++arms_remain; arms[j++] = i; }
    assert(j == arms_remain);

    if (arms_remain == 0 || n >= N) return 0;
    if (arms_remain == 1) {
	for (i=n+1; i<= N; i++) 
	  if (assign[i-1]==k) assign[i] = arms[0];
	return arms_remain;
    }
    for (i=n+1; i <= N; i++) 
      if (assign[i-1]==k) ++rest; /* #patients who need reassignment */
    if (rest == 0) return 0;
    sample = (double*) malloc(rest * sizeof(double));
    index = (int*) malloc(rest * sizeof(double));
    if (!sample || !index) { 
	printf("** Cannot re-randomize.\n"); 
	exit(0); }
    for (i=0; i<rest; i++) {
	sample[i] = (double)genunf((float)0.0, (float)1.0);
	index[i] = (i % arms_remain);
    }
    qrank(sample, index, 0, rest-1);
    j = 0;
    for (i=n+1; i<= N; i++) /* re_assignment */
      if (assign[i-1] == k) assign[i-1] = arms[index[j++]];
    free(sample); free(index);
    return arms_remain;
}/*------------------------------end of re_assign()----------------------*/


/* flag reassign = 1: reassign when one arm terminated */
void randomized_trial(Design *design, int *reassign, int *extended)
{
/*------------------------------------------------------------
  reassign : action after one arm is terminated
  extended : option for comparing two arms
  ------------------------------------------------------------*/
    int i, j, k, t, hit[10], *N, *assign, *select_cnt, *stop_cnt, best; 
    int nmax1, nmax2;
    double *postmu, max, min;   
    /* int size[TRIALSIZE][11];   */  
    /* this is too big for PC. Need better implementation */ 
    Event *comp, *cond;  /* int quantile[5], complete_cnt; */
    FILE *fp; char *name, buf[200];

    static int flag;
    
    int **result, tot, arms; Event *target;

    arms = design->Nprob; 
    target = design->current_event;
    tot =  design->number_of_event + design->number_of_cond_event;
    result = imatrix(TRIALSIZE, tot);
    /* poisson(mu) */
    /*  mu = design->accrual_rate * design->monitor_period; */
    /* N[j] = # patient for treatment j */
    N = (int*) malloc(arms * sizeof(int));
    postmu = (double*) malloc(arms * sizeof(double));
    select_cnt = (int*) malloc((arms+1) * sizeof(int));
    stop_cnt = (int*) malloc((arms+1) * sizeof(int));
    if (!result || !N || !postmu || !select_cnt || !stop_cnt) {
	printf("** Cannot allocate memory in Randomized trial.\n"); 
	exit(0);
    }
    /* initialize global counters before the simulation */
    for (j = 0; j <= arms; j++) { /* select_cnt[0] = #(select none)    */
	select_cnt[j] = 0;        /* select_cnt[j+1] = #(E_j selected) */
	stop_cnt[j] = 0;          /* stop_cnt[j] = #(E_j terminated)   */
    }

    /*----------------------------------------------------------*/
    for (i = 0; i < TRIALSIZE; i++)  {/* the i_th trial */
	/* initialization */
	int remain = design->Nprob; /* number of arms */
	/*  if (design->monitor_method == 2) *//* Poisson monitoring */
	/*  checkpoint =  (int) ignpoi( (float) mu ); 	*/  
	for (j = 0; j <= arms; j++) {
	    for (comp = design->event0->next; comp!=NULL; comp = comp->next){
		comp->outcome[j] = 0;
		for (cond = comp->child; cond != NULL; cond = cond->next)
		  cond->outcome[j] = 0;
	    }
	    hit[j] = 0; N[j] = 0; 
	} 
	/* generate randomization scheme */
	assign = balrand(arms, design->Nmax);

	if ( MERGE && arms == 2) {
	/*  make the 1st arm : 2nd arm = 2 : 1   */
	    for (j = 0; j < design->Nmax; j++) 
		if (assign[j] == arms) assign[j] = arms - 1;
	    nmax2 = design->Nmax / 3;
	    nmax1 = 2 * nmax2; 
	    /* printf("nmax1 = %d, nmax2 = %d\n", nmax1, nmax2); */
	}
 
	if (assign == NULL) {
	    printf("** balrand() fails.\n");
	    exit(0);
	}

	for (j = 1; j < design->Nmax; j++) { /* j_th patient */
	    k = assign[j-1];/* assign patient j treatment k = assign[j] */
	    if(hit[k] != 0) continue; 
	    ++N[k];           /* update #patients in treatment arm k */
	    /* simulate one patient -> update count x_k */
	    run_trial(k, design, design->trueprob[k]); 
	    
	    if ( j < design->Nmax ) {  
		if ( j % design->cohort_size == 0 )
		    hit[k] = check_bound(N[k], k, design, result[i]); 
	    }
	    if (hit[k] != 0) { /* arm k terminated */
		if(*reassign != 0) 
		    remain = re_assign(k, arms, j, design->Nmax, assign, hit);
		else --remain;
		if (remain == 0) break; 
		/* all arms terminated or nobody need reassignment */
	    } 
	}/*--------end of j_th patient */

	/* size[i][0] = j;   /* overall sample size */
	k = assign[design->Nmax-1];   /* the last patient */
	if (hit[k] == NO) ++N[k];
	for (k = 0; k < arms; k++) { 
	    /* size[i][k+1] = N[k];  */
	    if (hit[k] != NO) ++stop_cnt[k+1]; 
	}
	if (j < design->Nmax) { /* the trial stopped early */
	    ++select_cnt[0]; /* no selection */
	    ++stop_cnt[0];
	} 
	else {              /* the trial is complete */
	    result[i][0] = design->Nmax;
	    /* ++complete_cnt; */
	    /* select the best treatment E_k */
	    
	    for (t = 0; t < arms; t++) { 
		postmu[t] = (hit[t] != NO)?  -1.0 :
		    (double)( target->aE + target->outcome[t] )/
		    ( target->aE + target->bE + N[t] );
            }

	    if (target->type == EFFICACY) {  
  		max = 0.0; 
                for (t = 0; t < arms; t++) { 
                    if (hit[t] != NO) continue; /* skip the terminated arm */
                    if ( postmu[t] > max ) { 
                        max = postmu[t];
                        best = t; 
                    }
                    /* tie breaker */
                    if ( postmu[t] == max && genunf(.0, 1.0) <= .5 ) {
                        best = t;                        
                    }
                }
	    } 
            else if (target->type == ADVERSE) {
		min = 9999999.0;
		for (t = 0; t < arms; t++) { 
                    if (hit[t] != NO) continue; /* skip the terminated arm */
		    if ( postmu[t] >= .0 && postmu[t] < min ) {
			min = postmu[t];
			best = t;		
		    }
                    /* tie breaker */
                    if ( postmu[t] == min && genunf(.0, 1.0) <= .5 ) {
                        best = t;                        
                    }
		}                
	    }
	    ++select_cnt[best+1];
 	} 
	/* assert(result[i][0] == size[i][0]); */  /* check it later */
    /*-----------------------end of i_th trial ---------------------*/
    }

    /* tabulate the results:
       number of select E_j selected = select[j+1]; */
    sprintf(buf, "%s", design->filename);
    name = strcat(buf, ".simout3\0");
    if (!(fp = fopen(name, "a"))) {
	printf("cannot open file %s for storing simulation result.\n", name);
	exit(0);
    }
    printf("\nThe results of randomized trial\n"
	   "Criterion event: \"%s\"\n", target->name);
    printf("Event type: %s\n", (target->type != 0)? "Efficacy": "Adverse");
    printf("Re-assignment: %s\n", (*reassign != 0)? "yes" : "no");
    if(flag == 0) {  /* print overhead */
	if (!(fp = fopen(name, "w"))) {
	    printf("cannot open file %s for storing simulation result.\n", 
		   name);
	    exit(0);
	}
	fprintf(fp,"===================================================\n"
		"Randomized Trials Results For design:\n\"%s\"\n", 
		design->filename);
	fprintf(fp,"===================================================\n");
	flag = 1;   	(void) fclose(fp);
    }
    if (!(fp = fopen(name, "a"))) {
	printf("cannot open file %s for storing simulation result.\n", name);
	exit(0);
    }
    fprintf(fp, "\nCriterion event: \"%s\"\n", target->name);
    fprintf(fp, "Event type: %s\n",(target->type != 0)? "Efficacy": "Adverse");
    fprintf(fp, "Re-assignment: %s\n", (*reassign == YES)? "yes" : "no");
    if (*reassign == YES)
      fprintf(fp, "Sample size per arm: Nmin = %d, Nmax = %d x %d = %d\n",
	      design->Nmin, design->Nmax/arms, arms, design->Nmax);
    else
      fprintf(fp, "Sample size per arm: Nmin = %d, Nmax = %d\n",
	      design->Nmin, design->Nmax/arms);

    printf("\n\tTrue p-vector\n\t");
    fprintf(fp, "\n\tTrue p-vectors\n\t");
    for (j=0; j<design->number_of_element; j++) 
      printf("A_%d\t", j+1), fprintf(fp, "A_%d\t", j+1);
    printf("\n"); fprintf(fp, "\n");
    for (j = 0; j < arms; j++) {
	printf("E_%d\t", j+1);  fprintf(fp, "E_%d\t", j+1);
	for (k = 0; k < design->number_of_element; k++) {
	    printf("%.3f\t", design->trueprob[j][k]); 
	    fprintf(fp, "%.3f\t", design->trueprob[j][k]) ; 
	}
	printf("\n"); fprintf(fp, "\n");
    }
    printf("----------------------------------\n");
    fprintf(fp, "----------------------------------\n");

    printf("Prob(select E_j)\n");
    fprintf(fp, "Prob(select E_j)\n");
    for(j = 1; j <= arms; j++) 
      printf("E_%d   \t", j), fprintf(fp,"E_%d   \t", j);
    printf("None\n"); fprintf(fp, "None\n");
    printf("----------------------------------------------\n");
    fprintf(fp, "----------------------------------------------\n");
    for (j = 1; j <= arms; j++) 
      printf("%.4f\t", (double)select_cnt[j]/TRIALSIZE),
      fprintf(fp, "%.4f\t", (double)select_cnt[j]/TRIALSIZE);
    printf("%.4f\n", (double)select_cnt[0]/TRIALSIZE);
    fprintf(fp, "%.4f\n", (double)select_cnt[0]/TRIALSIZE);
    printf("----------------------------------------------\n");
    fprintf(fp, "----------------------------------------------\n"); 
  
    printf("\n** Save randomized trial result in file:\n\"%s\"\n", name);
    (void) fclose(fp);
    free(result);
    free(assign); free(N); free(postmu); 
    free(select_cnt); free(stop_cnt);
   
}/*----end of randomized_trial()--------*/




