/* "simulate.c" */
/* Time-stamp: <97/12/18 13:26:12 hgs> */
#include <stdio.h> 
#include <stdlib.h> 
#include <stddef.h>
#include <math.h> 
#include <string.h>
#include <assert.h>

#include "header.h"
#include "ranlib.h" 
  
int power(int, int);

/* runs the simulation */
void simulation(Design *design)
{
    int duration[TRIALSIZE]; /* for my own good */
    int i, tot;   int **result; /* int quantile[5]; */
    tot = design->number_of_event + design->number_of_cond_event;
   
    result = imatrix(TRIALSIZE, tot+1); 
   
    /* overhead for the output file */
    print_overhead(design); 

    for (i=0; i < design->Nprob; i++) { /* the i_th simulation */
	printf("\nRunning Simulation #%d, %d more to go..\r", 
	       i+1, design->Nprob - i - 1);

	simulate(design, design->trueprob[i], result, duration);
	print_primary_result_in_file(design, design->trueprob[i], result); 
	print_secondary_result_in_file(design, design->trueprob[i], result);
	/* report the duration. have to do it anyway */
	/* qsort(duration, TRIALSIZE, sizeof(int), intcompare);
	   quantile[0] = duration[(int)(TRIALSIZE* .10)];
	   quantile[1] = duration[(int)(TRIALSIZE* .25)]; 
	   quantile[2] = duration[(int)(TRIALSIZE* .50)];
	   quantile[3] = duration[(int)(TRIALSIZE* .75)];
	   quantile[4] = duration[(int)(TRIALSIZE* .90)];
	   printf("**** duration: ( ");
	   for (j = 0; j<=4; j++)  printf("%3d ", quantile[j]);
	   printf(")\n"); */
    }
    
    free(result);
}/*--------------------------------------------end of simulation()------*/

/* result of the k_th simulation.  output file "filename.simout1"  */
/* summarize the simulation result, compute operational characteristics */
/* current implementation: compute it, not store it */
void print_primary_result_in_file(Design *design, double *trueprob, 
				  int **result)
{
    /* overall prob(STOP) and  achieved sample size */
    int i;  int temp[TRIALSIZE];   double p_stop;  
    FILE *fp;  char buf[200], *name;
    int quantile[5];  
    int complete_cnt;

    sprintf(buf, "%s", design->filename);
    name = strcat(buf, ".simout1\0");
    
    fp=fopen(name, "a");
    if (fp == NULL) {
	printf("cannot open file %s for storing simulation result.\n", name);
	exit(0);
    }
    /* prob(STOP) */
    complete_cnt = 0;
    for (i=0; i < TRIALSIZE; i++) {
	temp[i] = result[i][0];  /* sizeof(trial i) when stopped */
	if (result[i][0] == design->Nmax) ++complete_cnt;  
    }
    p_stop = (double) (TRIALSIZE - complete_cnt)/TRIALSIZE;
    
    /* sort the array of trial size */
    qsort(temp, TRIALSIZE, sizeof(int), intcompare); 

    quantile[0] = temp[(int)(TRIALSIZE* .10)];
    quantile[1] = temp[(int)(TRIALSIZE* .25)]; 
    quantile[2] = temp[(int)(TRIALSIZE* .50)];
    quantile[3] = temp[(int)(TRIALSIZE* .75)];
    quantile[4] = temp[(int)(TRIALSIZE* .90)];
    
    fprintf(fp, "( ");
    for (i = 0; i < design->number_of_element - 1; i++) 
      fprintf(fp, "%.3f,", trueprob[i]);
    fprintf(fp, "%.3f )", trueprob[i]);
    
    fprintf(fp, "\t%.4f\t\t", p_stop); 
    
    fprintf(fp, "( ");
    for (i = 0; i<=4; i++) 
      fprintf(fp, "%3d ", quantile[i]);
    fprintf(fp, ")\n");
        
    (void)fclose(fp);
}/*--------------------end of print_primary_result_in_file()--------------*/


/* secondary simulation result */
/* pretty complicated. fix it later */
/* try KISS approach */
void print_secondary_result_in_file(Design *design, double *trueprob, 
				    int **result)
{
    int i, j, k, ncol, num_event, cell;  char buf[200], *name; FILE *fp;
    int *count;  Event *comp, *cond; int cell2, *count2;
    int *bnd_type;
    num_event = design->number_of_event + design->number_of_cond_event;
    bnd_type = (int*) malloc(num_event * sizeof(int));
    /* this implementation is ok if num_event <= 16. this should be ample. */
    k = power(2, num_event); 
    ncol = 2 * num_event;
    count = (int*) malloc( k * sizeof(int) ); 
    count2 = (int*) malloc( k * k * sizeof(int) );
    if (count == NULL || count2 == NULL || bnd_type == NULL ) { 
	printf("** Not enough memory for break-down counting.\n");
	exit(0);
    }
    for (i=0; i < k; i++) count[i] = 0;
    for (i=0; i < k*k; i++) count2[i] = 0;
    sprintf(buf, "%s", design->filename);
    name = strcat(buf, ".simout2\0");
    
    if (!(fp=fopen(name, "a"))) {
	printf("cannot open file %s for storing simulation result.\n",
	       name);
	exit(0);
    } 
    /* real counting action */
    for (j=0; j < TRIALSIZE; j++) {
	cell = 0; cell2 = 0;
	for (i=0; i < num_event; i++) {
	    if ( result[j][i+1] !=  0 )  cell += power(2,i);
	    if ( result[j][i+1] == -1 ) cell2 += power(2, 2*i);
	    if ( result[j][i+1] ==  1 ) cell2 += power(2, 2*i+1);
	}
	++count[cell]; 	
	++count2[cell2];
    }
    /* report prob(stop) , old implementation doesn't report upper/lower */
    /*    fprintf(fp, "--------------------------------------\n");
	  for (comp=design->event0->next; comp!=NULL; comp=comp->next) 
	  fprintf(fp, "%s\t", comp->name);
	  for (comp=design->event0->next; comp!=NULL; comp=comp->next) 
	  for (cond=comp->child; cond != NULL; cond=cond->next) 
	  fprintf(fp, "%s\t", cond->name);
	  fprintf(fp, "prob(stop)\n");
	  fprintf(fp, "--------------------------------------\n");
	  
	  for (i=0; i < power(2,num_event); i++) {
	  for (j=0; j < num_event; j++) 
	  fprintf(fp, "%s\t", ((i >> j) & 1)? "stop": "no  ");
	  fprintf(fp, "%.4f\n", (float) count[i]/TRIALSIZE); 
	  }    
	  fprintf(fp, "--------------------------------------\n");
	  fprintf(fp, "Overall Prob(stop) = %.4f\n", 
	  1.0 - (float)count[0]/TRIALSIZE); 
	  fprintf(fp, "--------------------------------------\n\n");
	  */
 /* new implementation: indicate hit.lower & hit.upper */
    fprintf(fp, "-------------------------------------------------------\n");
    fprintf(fp, "true prob: (");
    for (i = 0; i < design->number_of_element - 1; i++) 
      fprintf(fp, "%.3f, ", trueprob[i]);
    fprintf(fp, "%.3f )\n", trueprob[i]);
    /* print the names of events */
    fprintf(fp, "-------------------------------------------------------\n");
    for (comp=design->event0->next; comp!=NULL; comp=comp->next) {
	if (comp->boundary_type != 4) fprintf(fp, "%s\t", comp->name); 
	if (comp->boundary_type == 2) fprintf(fp, "\t");
    }
    for (comp=design->event0->next; comp!=NULL; comp=comp->next) 
      for (cond=comp->child; cond != NULL; cond=cond->next) {
	  if (cond->boundary_type != 4) fprintf(fp, "%s\t", cond->name);   
	  if (cond->boundary_type == 2) fprintf(fp, "\t");
      }
    fprintf(fp, "\n");    
    /* indicate hit.lower & hit.upper */
    i = 0;
    for (comp=design->event0->next; comp!=NULL; comp=comp->next) { 
	bnd_type[i++] = comp->boundary_type;
	if (comp->boundary_type == 1) 
	  fprintf(fp, "lower\t");
	else if (comp->boundary_type == 3)
	  fprintf(fp, "upper\t");
	else if (comp->boundary_type == 2)
	  fprintf(fp, "lower/upper\t");
	/* else  fprintf(fp, "no bound\t"); */
    }
    for (comp=design->event0->next; comp!=NULL; comp=comp->next) 
      for (cond=comp->child; cond != NULL; cond=cond->next) {
	  bnd_type[i++] = cond->boundary_type;
	  if (cond->boundary_type == 1) 
	    fprintf(fp, "lower\t");
	  else if (cond->boundary_type == 3)
	    fprintf(fp, "upper\t");
	  else if (cond->boundary_type == 2)	    
	    fprintf(fp, "lower/upper\t");
	  else  fprintf(fp, "no bounds\t");
      } 
    fprintf(fp, "\tprob(stop)\n");
    fprintf(fp, "-------------------------------------------------------\n");
    
    for (i=0; i < power(2, ncol); i++) {
	if (count2[i] == 0 && i != 0) continue;
	/* for (j=0; j < ncol; j++) 
	  fprintf(fp, "%s\t", ((i >> j) & 1)? "stop": "no  "); */
	for (j=0; j < num_event; j++) {
	    if (bnd_type[j] == 1) {
		fprintf(fp, "%s\t", ((i >> (2*j)) & 1)? "stop":   "no  ");
	    } 
	    else if (bnd_type[j] == 2) {
		fprintf(fp, "%s/", ((i >> (2*j)) & 1)? "stop":    "no  ");
		fprintf(fp, "%s\t", ((i >> (2*j+1)) & 1)? "stop": "no  ");
	    }
	    else if (bnd_type[j] == 3) {
		fprintf(fp, "%s\t", ((i >> (2*j+1)) & 1)? "stop": "no  ");
	    }
	}
	fprintf(fp, "\t%.4f\n", (float) count2[i]/TRIALSIZE); 
    }    
    fprintf(fp, "-------------------------------------------------------\n");
 /*   fprintf(fp, "Marginal:\n");  */
    fprintf(fp,"Overall prob(stop) = %.4f\n", 1.0 - (double)count2[0]/TRIALSIZE);
    
    fprintf(fp, "-------------------------------------------------------\n\n");
    (void)fclose(fp);
    free(bnd_type); free(count); free(count2);
}/*----------------------end of print_secondary_result_in_file()-------*/


int power(int base, int j)
{
    int i; int ans = 1;
    for (i=0; i<j; i++) ans *= base;
    return ans;
}


void 
simulate(Design *design, double *trueprob, int **result, int *duration)
{
/* simulate trials for given true prob of each elementary event    
   design: pre-specified parameters of each trial                    
   result[i][0] is the final patient number when the i_th trial stop 
   result[i][j] = 0: doesn't hit boundary of event j
                  1: hit event j's upper bound
		 -1: hit event j's lower bound                     
   result is (trialsize x (event_count+1))

   workhorses:   run_trial() and check_bound()
*/   
    int i, j, k, hit, tot, checkpoint; Event *comp, *cond;
    double mu, period, window, first_period; /* int duration;*/
    /* when X ~ Poisson can use Barry Brown's ranlib.c 
       This should not affect the way the count x[i] is generated
       but the time to check bounds, of course, in check_bound().
       Do this later.  */
     /* initiate the matrix result[][] */
    tot = design->number_of_event + design->number_of_cond_event;
    for (i = 0; i < TRIALSIZE; i++)  for (j=0; j <= tot; j++)   
      result[i][j] = 0; 

    /* poisson(mu) */
    mu = design->accrual_rate * design->monitor_period; 
    /* for response window > 0, the only difference is:
       the first period = k * period - window > 0 */
    window = design->response_window;
    period = design->monitor_period;
    /* start the trial */
    for (i = 0; i < TRIALSIZE; i++) {/* one loop, one trial */
	if (design->monitor_method == 2) { /* Poisson monitoring */
	    duration[i] = 0;
	    if (window > 0.0) { /* response window > 0 */
		first_period = period - window;
		++duration[i];
		while (first_period < 0.0) { 
		    first_period += period;
		    ++duration[i];
		}
		checkpoint = 
		  (int)ignpoi((float)(first_period * design->accrual_rate));
		while (checkpoint == 0) {
		    checkpoint = (int) ignpoi( (float) mu );
		    ++duration[i];
		}
	    } 
	    else {  /* response window = 0 */
		do { checkpoint = (int) ignpoi( (float) mu ); 
		     ++duration[i];
		 } while (checkpoint == 0); 
	    }
	}
	  /* initialize the outcome counts for each events */
	for (comp = design->event0->next; comp != NULL; comp = comp->next){
	    comp->out = 0; comp->outcome[0] = 0;
	    for (cond = comp->child; cond != NULL; cond = cond->next) {
	      cond->out = 0;  cond->outcome[0] = 0; }
	}
	/* accrual the j_th patient  --------------------------------*/
	for (j = 1; j < design->Nmax; j++) {
	    run_trial(0, design, trueprob); 	    
	    hit = 0;
	    if ( j < design->Nmax) {  
		if (design->monitor_method == 1) { /* monitor by cohort */
		    if ( j % design->cohort_size == 0 )
		      hit = check_bound(j, 0, design, result[i]); 
		}
		else { /* Poisson monitoring */
		    if ( j == checkpoint ) { 
			hit = check_bound(j, 0, design, result[i]); 
			do {
			    k = (int) ignpoi((float)mu);
			    ++duration[i];
			} while ( k == 0 );
			checkpoint += k;
		    } 
		}
	    }
	    /* when hit(!=0) the boundary, record the number of patients, 
	       event, boundary type and go to next trial */
	    if (hit != 0) break; 
	}
	if (j == design->Nmax) result[i][0] = design->Nmax;
    }    
}/*----------------------------- End of simulate() ----------------------*/


void 
run_trial(int arm, Design *design, double *trueprob) 
{
/*
   generate one elementary outcome and update the (binary) outcome for each 
   compound and conditional events on treatment arm    */

    int k, element;  double dice;  Event *comp, *cond; 
 
    dice = (double) genunf((float)0.0, (float)1.0); 
    element = design->number_of_element - 1;
    for (k=0; k < design->number_of_element; k++) { 
	dice -= trueprob[k];
	if (dice <= 0.0)  {
	    element = k;
	    break; /* the elementary event k occurred */
	}
    }
    /* compound events */
    for (comp = design->event0->next; comp != NULL; comp = comp->next) {
	if (comp->definition[element] == 1) { 
	    ++comp->out; ++comp->outcome[arm]; 
	}
	/* conditional events */
	for (cond = comp->child; cond != NULL; cond = cond->next)
	  if (cond->definition[element] == 1) { 
	      ++cond->out;  ++cond->outcome[arm];
	  }
    }
}/*------- End of run_trial() ----------------------------------------*/



int 
check_bound(int n, int arm, Design *design, int *result)
{
/* called by simulate() to check boundary of every event at total
   patients = n.  return 0 if not hit any boundary.  */
/*------------------------------------------------------------------
  argument list:
    result : result[0] = total number of patient n
             for k-th trial:    
             result[k] = -1: hit lower bound; 
                          0: no hit; 
                          1: hit upper bound;
**Remark: should include  2: hit both bound;			  
  ------------------------------------------------------------------*/
    int m, a, b, lobnd, upbnd; Event *comp, *cond;
    int hit = 0;    int event_index = 0;
    
    /* compound events */
    for (comp = design->event0->next; comp != NULL; comp = comp->next) {
	++event_index;
	if (comp->boundary_type == 4) continue; /* skip this event */

	if (n < design->Nmin) { /* "semi-free-ride" zone */ 
	    a = comp->lowerbound[design->Nmin];
	    b = comp->upperbound[design->Nmin];

	    lobnd = n + a - design->Nmin;  /* negative means free ride */
	    upbnd = b;  /* b > n means free ride */
	}
	else { 
	    lobnd = comp->lowerbound[n];
	    upbnd = comp->upperbound[n];
	}
	if (comp->boundary_type == 1 || comp->boundary_type == 2) {
	    if ( comp->outcome[arm] <= lobnd ) { 
		result[0] = n;
		result[event_index] = hit = -1;
	    }
	}
	if (comp->boundary_type == 3 || comp->boundary_type == 2) {
	    if ( comp->outcome[arm] >= upbnd ) { 
		result[0] = n;
		result[event_index] = hit = 1; 
	    }
	}
	/* should include hit both bounds , fix it later */
    }    
    /* conditional events */    /* very tricky! */
    for (comp = design->event0->next; comp != NULL; comp = comp->next)
      for (cond = comp->child; cond != NULL; cond = cond->next) {
	  ++event_index;	 
	  if (cond->boundary_type == 4) continue; /* skip this event */
	  m = cond->parent->outcome[arm]; 
	  /* m = cond->parent->out; */
	  /* a "local" Nmin will apply later */
	  /* if (m < design->Nmin) continue; */
	  /* if (m < design->Nmin) {*/  /* "free-ride" zone */ 
	  /*     a = cond->lowerbound[design->Nmin];
		 b = cond->upperbound[design->Nmin];
		 lobnd = n + a - design->Nmin;
		 upbnd = b; 
		 } else {  
		 lobnd = cond->lowerbound[m];
		 upbnd = cond->upperbound[m];
		 } */
	  lobnd = cond->lowerbound[m];
	  upbnd = cond->upperbound[m];
	  if (cond->boundary_type == 1 || cond->boundary_type == 2) {
	      if (cond->outcome[arm] <= lobnd) { 
		  result[0] = n;
		  result[event_index] = hit = -1;
	      }
	  }
	  if (cond->boundary_type == 3 || cond->boundary_type == 2) {
	      if (cond->outcome[arm] >= upbnd){
		  result[0] = n;
		  result[event_index] = hit = 1; 
	      }	  
	  }
      }    
    assert(event_index == design->number_of_event +
	   design->number_of_cond_event);
    return hit;
/*----------------------------------------end of check_bound()-----*/
}




/* define for qsort() */
int intcompare(const void *a, const void *b)
{
    int x = (int) *(int*) a;
    int y = (int) *(int*) b;
    return (x - y);
}


void 
print_overhead(Design *design)
{
/* this stupid function exists because we want to append new
   simulation result again and again to the file */
/* stupid overhead of the output file 
   flag = 1: primary; flag = 2: secondary */
    int i;
    FILE *fp; char buf[200], *name1, *name2;

    /* primary result */
    sprintf(buf, "%s", design->filename);
    name1 = strcat(buf, ".simout1\0");
    
    if (!(fp=fopen(name1, "w"))) {
	printf("cannot open file \"%s\" properly!\n",name1);
	exit(0);
    }	
    printf("\n** Save simulation results in file:\n\"%s\"\n\n", name1);
    fprintf(fp,"\n");

    /* avoid double print */
    /* st = get_file(name1); 
       if ((int)strlen(st) > 10) {  (void)fclose(fp); return; } */

    fprintf(fp,"===================================================\n"
	    "Simulation Results For design:\n\"%s\"\n", design->filename);
    fprintf(fp,"===================================================\n\n");
    
    fprintf(fp,"Number of Repetitions : %d\n\n", TRIALSIZE); 
    
    fprintf(fp, "\"True\" Probability Vector\t\t"
	    "   Achieved Sample Size Percentiles\n");
    
    fprintf(fp, "(");
    for (i=0; i < design->number_of_element-1; i++)
      fprintf(fp, "   p%d,", i+1);
    fprintf(fp, "   p%d )", design->number_of_element);
    fprintf(fp,"\tPr(STOP)\t   10  25  50  75  90\n\n");
    
    (void)fclose(fp);
  
    /* secondary result */
    sprintf(buf, "%s", design->filename);
    name2 = strcat(buf, ".simout2\0");
    
    if (!(fp=fopen(name2, "w"))) {
	printf("cannot open file \"%s\" properly!\n",name2);
	exit(0);
    }	
    printf("** Save the additional simulation results in file:\n"
	   "\"%s\"\n", name2);
    fprintf(fp,"\n===================================================\n"
	    "Simulation Results For design:\n\"%s\"\n", design->filename);
    fprintf(fp, "-- A table broken down by Reason for Stopping\n"
	    "===================================================\n\n");
    (void)fclose(fp);
 /*---------end of print_overhead()-----------------------*/
}
