/* "struct.c" */
/* Time-stamp: <97/05/14 17:53:50 hgs> */
/* data structures operations */

#include <stdio.h> 
#include <stdlib.h> 
#include <stddef.h>
#include <math.h> 
#include <string.h>
#include <assert.h>

#include "header.h"

extern int MIX_K;
extern int NMIN, NMAX, NUMBER_OF_ELEMENT;
/* initialize the template : need a global object as the template */
/* default setup */
void initial_design(Design *design, int Nmin, int Nmax,  
		    int number_of_element, char *file)
{
    int i;
    design->Nmin = Nmin;
    design->Nmax = Nmax;
    design->number_of_element = number_of_element;
    design->number_of_event = 0;
    design->number_of_cond_event = 0;
    design->aS = (double*) malloc(number_of_element * sizeof(double));
    design->aE = (double*) malloc(number_of_element * sizeof(double));
    design->muS = (double*) malloc(number_of_element * sizeof(double));
    design->muE = (double*) malloc(number_of_element * sizeof(double)); 
    design->event0 = (Event*) malloc(sizeof(Event));
    /* malloc() fails */
    if (!design->aS || !design->aE || !design->muS 
	|| !design->muE || !design->event0) {
	design = NULL; 
	return;
    }
    /* default aS, aE */
    for(i=0; i< number_of_element; i++) {
	design->aS[i] = 1.0; 
	design->aE[i] = 1.0;
    }    
    /* parameters of monitoring: default */
    design->cohort_size = 1;
    design->monitor_method = 1;   /* default is cohort monitoring */
    design->monitor_period = 0.0;   /* for Poisson accrual */
    design->accrual_rate = 0.0;

    /* for simulation */
    /*   design->trueprob = double[20][50] */
    design->Nprob = 0;
    /* The list of compound events. need add_event() to fill in the list */
    strcpy(design->event0->name, "head of event list");
    design->event0->type = -1;   /* not adverse nor efficate */
    design->event0->next = NULL; /* end of the list */
    /*    design->event0->last = NULL; */
    design->event0->child = NULL;
    design->event0->design = design;
    design->current_event = design->event0;
 
    strcpy(design->filename, file);   /* output file name */ 
} /*-------------------- End of initial_design() -------------------*/


/* retrieve old design from file "name", assign pointer *design to it 
   *design must be initialized before retrieving */
int retrieve_old_design(char *name, Design *design)
{
    int i, k;  Event *event;
    char *text, *st, *st2;  char buf1[200], buf2[200]; char *temp;
    text = get_file(name);
    temp = (char*) malloc(200*sizeof(char));
    if ( text == NULL || temp == NULL) {
	printf("** Cannot retrieve the file:\n\"%s\".\n", name);
	return FAIL;
    } 
    if ( (int)strlen(text) < 10 ) {
	free(text);  free(temp);
	return FAIL;
    }
    printf("\n** Retrieveing design from file:\n\"%s\"\n....\n", name);
    /* retrieveing data */    
    if (get_data(text, "Design Name:", temp) == FAIL) return FAIL;
    strcpy(design->filename, temp);  
    if (get_data(text, "Nmin =", temp) == FAIL) return FAIL;
    NMIN = atoi(temp);
    if (get_data(text, "Nmax =", temp) == FAIL) return FAIL;
    NMAX = atoi(temp);
    if (get_data(text, "Number of elementary events =", temp) == FAIL)
      return FAIL;
    NUMBER_OF_ELEMENT = atoi(temp);

    design->Nmin = NMIN;
    design->Nmax = NMAX;
    design->number_of_element = NUMBER_OF_ELEMENT;
    initial_design(design, NMIN, NMAX, NUMBER_OF_ELEMENT, design->filename);
    if (design == NULL) { 
	printf("** Not enough memory for retrieving old design.\n");
	return FAIL;
    }
    design->number_of_event = 0;
    design->number_of_cond_event = 0;
    if (get_data(text, "Monitoring method =", temp)==FAIL) return FAIL;
    if ( strcmp(temp, "By cohort") == 0) design->monitor_method = 1;
    else design->monitor_method = 2;

    if(design->monitor_method == 1) {
	if (get_data(text, "Cohort size =", temp)==FAIL) return FAIL;
	design->cohort_size = atoi(temp);
    }
    if(design->monitor_method == 2) {
	if (get_data(text, "Accrual rate =", temp)==FAIL) return FAIL;
	design->accrual_rate = atof(temp);
	if (get_data(text, "Monitoring period =", temp)==FAIL) return FAIL;
	design->monitor_period =  atof(temp);
	if (get_data(text, "Response window =", temp)==FAIL) return FAIL;
	design->response_window =  atof(temp);
    }
    if (get_data(text, "Standard Prior Type:", temp)==FAIL) return FAIL;
    /* temp = get_data(text, "Standard Prior Type:"); */
    if( strcmp(temp, "Mixture") == 0 ) {
	int i, j;
	if (get_data(text, "Number of Mixture =", temp)==FAIL) return FAIL;
	MIX_K = atoi(temp);
	design->Sprior_type = MIX_K;

	for(i = 0; i < MIX_K; i++) {
	    char aa[20], buf[10];
	    sprintf(buf, "Dir_%d", i+1);
	    if (get_data(text, buf, temp)==FAIL) return FAIL;
	    /* temp = get_data(text, buf); */
	    for(j = 0; j < design->number_of_element; j++) {
		sscanf(temp, "%s", aa);
		design->mix_aS[i][j] = atof(aa);
		temp += strlen(aa) + 1;
	    }
	    sscanf(temp, "%s", aa);
	    design->Sprior_weight[i] = atof(aa);
	    temp += strlen(aa);
	}
	for(i=0; i < design->number_of_element; i++) {
	    sprintf(buf1, "a_E%d =", i+1);
	    if (get_data(text, buf1, temp)==FAIL) return FAIL;
	    design->aE[i] = atof(temp); /*atof(get_data(text, buf1));*/
	}
    }
    else  {  /* single prior */ 
	design->Sprior_type = MIX_K = 1; 	
	for(i=0; i<design->number_of_element; i++) {
	    sprintf(buf1, "a_S%d =", i+1);
	    sprintf(buf2, "a_E%d =", i+1);
	    if (get_data(text, buf1, temp)==FAIL) return FAIL;
	    design->aS[i] = atof(temp); /* atof(get_data(text, buf1));*/
	    if (get_data(text, buf2, temp)==FAIL) return FAIL;
	    design->aE[i] = atof(temp); /* atof(get_data(text, buf2));*/
	}
    }
    /* retrieve events list */ /* tricky */
    st = strstr(text, "Compound Events:");
    st2 = strstr(text, "Conditional Events:");
    k = strlen(st2);
    /* compound events */
    while ((st = strstr(st, "Event Name:")) != NULL) 	
      {
	  if ((int)strlen(st) < k)  
	    break;  /* move on to cond events */ 
	  event = retrieve_event(st);
	  add_event(event, design);
	  st += strlen("Event Name:");
      }
    /* conditional events */
    while (st != NULL)
      { 
	  retrieve_cond_event(st, design);  
	  st += strlen("Event Name:");
	  st = strstr(st, "Event Name:");
      }    
    compute_design_boundary(design);
    free(text); free(temp);
    return SUCCEED;
}/*--------------------------------end of retrieve_old_design()------*/


Event *retrieve_event(char *text)
{
    Event *event;  char buf[200], name[100], temp[200];
    int i, type, k;
  
    event = (Event*) malloc(sizeof(Event));
    if (event == NULL) return event;
    /* name */
    /* strcpy(name, get_data(text, "Event Name:")); */
    get_data(text, "Event Name:", temp);
    strcpy(name, temp);
    /* strcpy(buf, get_data(text, "Event Type:")); */
    get_data(text, "Event Type:", temp);
    strcpy(buf, temp); 
    type = (strcmp(buf, "Adverse")==0)?  0 : 
      (strcmp(buf, "Efficacy")==0)?  1 : -1;
    
    event = make_event(type, name, NMAX, NUMBER_OF_ELEMENT, 
		       NULL, NULL, NULL, .0, .0, .0, .0);
    /* definition */
    /* strcpy(buf, get_data(text, "Definition code:")); */
    get_data(text, "Definition code:", temp);
    strcpy(buf, temp);
    for(i = 0; i< NUMBER_OF_ELEMENT; i++)
      event->definition[i] = ( buf[i] == '1')?  1 : 0;

    if ( MIX_K > 1 ) { /* mixture prior */
	for(i = 0; i < MIX_K; i++) {
	    char aa[20], bb[20], ww[20], buf1[10];
	    sprintf(buf1, "Dir_%d", i+1);
	    /* temp = get_data(text, buf1); */
	    get_data(text, buf1, temp);
	    sscanf(temp, "%s %s %s", ww, aa, bb);
	    event->mix_aS[i] = atof(aa); 
	    event->mix_bS[i] = atof(bb);
	}	
    }
    else { /* single prior */
	get_data(text, "a_S =", temp);	event->aS = atof(temp);
	get_data(text, "b_S =", temp);	event->bS = atof(temp);
	get_data(text, "a_E =", temp);	event->aE = atof(temp);
	get_data(text, "b_E =", temp);	event->bE = atof(temp);
	/* event->aS = atof(get_data(text, "a_S ="));
	   event->bS = atof(get_data(text, "b_S ="));
	   event->aE = atof(get_data(text, "a_E ="));
	   event->bE = atof(get_data(text, "b_E =")); */
    }
    get_data(text, "boundary type =", temp);
    k = atoi(temp);
    event->boundary_type = k;
    if(k == 1 || k == 2) {
	get_data(text, "delta_L =", temp); event->deltaL = atof(temp);
	get_data(text,"p_L =", temp); event->pL = atof(temp);
	/* event->deltaL = atof(get_data(text, "delta_L ="));
	   event->pL = atof(get_data(text, "p_L =")); */
    }
    if(k == 2 || k == 3) {
	get_data(text, "delta_U =", temp); event->deltaU = atof(temp);
	get_data(text,"p_U =", temp); event->pU = atof(temp);
	/* event->deltaU = atof(get_data(text, "delta_U ="));
	   event->pU = atof(get_data(text, "p_U =")); */
    }
    return event;
}/*-------------------------------end of retrieve_event()-------------*/


/* retrieve conditional events from "text"  */
void retrieve_cond_event(char *text, Design *design)
{
    int i, element_cnt, *parent_def, *seed_def;  
    Event *comp, *comp1, *comp2;  char buf[100], buf1[100], temp[200];
    Event *cond_event;
    element_cnt = design->number_of_element;
    /* malloc() */
    cond_event = (Event*) malloc(sizeof(Event));
    parent_def = (int*) malloc(element_cnt * sizeof(int));
    seed_def =  (int*) malloc(element_cnt * sizeof(int));
    if (!cond_event || !parent_def || !seed_def) {
	cond_event = NULL; return;
    }
    cond_event = retrieve_event(text);
    /* finding the parent */
    get_data(text, "Parent code:", temp);
    strcpy(buf, temp);
    get_data(text, "Seed code:", temp);
    strcpy(buf1, temp);
    /* strcpy(buf, get_data(text, "Parent code:"));
       strcpy(buf1, get_data(text, "Seed code:")); */

    for(i=0; i < element_cnt; i++) {
	parent_def[i] = ( buf[i] == '1')?  1 : 0;
	seed_def[i] = ( buf1[i] == '1')? 1 : 0;
    }
    for(comp = design->event0->next; comp != NULL; comp = comp->next)
      { /* search for the parent */
	  for(i=0; i < element_cnt; i++) 
	    if (comp->definition[i] != parent_def[i])
	      break;
	  if (i == element_cnt) break;   /* comp is the parent */
      }
    for(comp1 = design->event0->next; comp1 != NULL; comp1 = comp1->next)
      { /* search for the seed */
	  for(i=0; i < element_cnt; i++)
	       if (comp1->definition[i] != seed_def[i])
		 break;
	  if (i == element_cnt) break;   /* comp1 is the seed */
      }
  
    /* legitimate child joins the family */
    /* a LIFO list */
    /*    cond_event->next = comp->child;
	  comp->child = cond_event;    */
    
    /* a FIFO list */
    if(comp->child == NULL) { 
	comp->child = cond_event;   
	cond_event->next = NULL;  
    }
    else {
	for(comp2=comp->child; comp2->next != NULL; comp2 = comp2->next) 
	  ;  
	comp2->next = cond_event;
	cond_event->next = NULL;
    } 
    
    cond_event->child = NULL;  
    cond_event->parent = comp;    /* cond_event is: (comp1 | comp) */
    cond_event->seed = comp1;     /* this is setup for remove_event() */
    cond_event->design = comp->design;

    ++comp->child_count;
    ++cond_event->design->number_of_cond_event; /* # of cond events */

    free(parent_def); free(seed_def);
}/*-------------------------------end of retrieve_cond_event()---------*/



/* design */
/* fill in the flesh and blood of a design */
Design *make_design(int Nmin, int Nmax, double *aS, 
		    double *aE, 
		    int number_of_element, 
		    int cohort_size, 
		    double monitor_period, double accrual_rate,
		    char *file)
{
    int i;
    Design *design = (Design*) malloc(sizeof(Design)); 
    design->Nmin = Nmin;
    design->Nmax = Nmax;
    design->number_of_element = number_of_element;
    design->number_of_event = 0;
    design->number_of_cond_event = 0;
    design->aS = (double*) malloc(number_of_element * sizeof(double));
    design->aE = (double*) malloc(number_of_element * sizeof(double));
    design->muS = (double*) malloc(number_of_element * sizeof(double));
    design->muE = (double*) malloc(number_of_element * sizeof(double));

    for(i=0; i< number_of_element; i++) {
	design->aS[i] = (aS == NULL) ?  0.0 : aS[i];  
	design->aE[i] = (aE == NULL) ?  0.0 : aE[i]; 
    }
    
    /* parameters of monitoring */
    design->cohort_size = cohort_size;
    design->monitor_period = monitor_period;   /* for Poisson accrual */
    design->accrual_rate = accrual_rate;

    /* The list of compound events. need add_event() to fill in the list */
    design->event0 = (Event*) malloc(sizeof(Event));
    strcpy(design->event0->name, "listhead: a null event");
    design->event0->next = NULL; /* head of the list */
    design->event0->design = design;
    /* array[] seems to be a better structure for design->eventlist.
       to cope with the upper limit of (# of event) can use realloc */
    /*   for(i=0; i< number_of_event; i++) 
	 design->event[i] = NULL;  */  /* default */
 
    strcpy(design->filename, file);   /* output file name */
 
    return design; 
} /*-------------------- End of make_design() -------------------*/


/* insert a new event after the head of the event list: design->event0 */
/* Notes: add_conditional_event is handled by make_cond_event() */
void add_event(Event *event, Design *design)
{
    Event *tail;
    if (event == NULL) { 
	printf("Illegal attempt to add NULL event.\n"); exit(1); 
    }
    /* FIFO list */  /* more natural in practice */
    for(tail = design->event0; tail->next != NULL; tail = tail->next)
      ;  /* reach the tail of the event list */
    tail->next = event;
    /*    event->last = tail;*/
    event->next = NULL;
    event->design = design;
    update_aSaE(event);
    ++design->number_of_event;
    design->number_of_cond_event += event->child_count;    
    design->current_event = event;
}/*----------------- End of add_event() -----------------------------*/


/* remove event from the event list */
void remove_event(Event *event, Design *design)
{
    Event *comp, *cond;
    if((event == design->event0) || (event == NULL))  return;
      /* cannot remove the head of event list and NULL event */
 
    if(event->parent == NULL) { /* compound event */
	/* First job: remove conditional events: (event|*) */
	for(comp = design->event0->next; comp != NULL; comp = comp->next)
	  for(cond = comp->child; cond != NULL; cond = cond->next)
	    if(cond->seed == event) { /* cond = (event|*) */
		remove_conditional_event(cond); 
		/* break; */ /* why? */
	    }
	/* then remove "event" itself */
	for(comp=design->event0; comp->next != event; comp= comp->next) 
	  ;  
	comp->next = event->next;
	--design->number_of_event;
	design->number_of_cond_event -= event->child_count;
    }
    else if(event->parent != NULL) { /* conditional event */
	/*   if(event->parent->child == event)
	     event->parent->child = event->next;
	     else  event->last->next = event->next;
	     --design->number_of_cond_event;*/
	remove_conditional_event(event);
    }
    /*    free_event(event); */
}/*----------------------------------------end of remove_event() ----*/

/* traverse event list and set design->current_event
   to the event chosen by the user */
void traverse_event_list(Design *design)
{
    Event *comp, *cond; int j, option, tot;
    tot = design->number_of_event + design->number_of_cond_event;
    if (tot == 0) { 
	printf("No event defined for design:\n\"%s\"\n", design->filename); 
	return; 
    }
    /* catalog */
    j = 0;
    printf("\n---------------------------------------------\n");
    printf("List of events for design:\n\"%s\"\n", design->filename);
    printf("---------------------------------------------\n");  
    /* compound events */
    for(comp = design->event0->next; comp != NULL; comp = comp->next){
	++j;
	printf("(%d) %s\n", j, comp->name);
    }
    assert(j == design->number_of_event);
    /* conditional events */
    for(comp = design->event0->next; comp != NULL; comp = comp->next)
       for(cond = comp->child; cond != NULL; cond = cond->next) {
	   ++j;
	   printf("(%d) %s\n", j, cond->name);
       }
    assert(j == tot);

    /* menu */
    do {
	printf("\nPick the event you want, or type 0 (zero) ");
	printf("to skip this menu.\n\n");
	printf("Your choice? \n> ");
	option = get_an_integer();
	if (option < 0 || option > tot) 
	  printf("\nMust choose between 0 and %d. Try again.\n", tot);
    } while (option < 0 || option > tot);
    /* compound events */
    j = 0;
    for(comp = design->event0->next; comp != NULL; comp = comp->next){
	++j;
	if ( j == option) {
	    printf("\nyou chose event:(%d) %s\n", j, comp->name);
	    design->current_event = comp;
	    return;
	}
    }
    /* conditional events */
    for(comp = design->event0->next; comp != NULL; comp = comp->next)
      for(cond = comp->child; cond != NULL; cond = cond->next) {
	 ++j;
	  if ( j == option) {
	      printf("\nyou choose event:(%d) %s\n", j, cond->name);
	      design->current_event = cond;
	      return;
	  }
      }
}/*----------------------------------end of traverse_event_list()----*/

void free_design(Design *design)
{
    if (!design) {
	free(design->aS);  	free(design->aE);
        free(design->muS);        free(design->muE);
	free_event(design->event0); 
	free(design);
    }
}/*----------------- End of free_design() ---------------------------*/


/* Event */
/* compound event can exist without link to design */
/* make_event(Design*,..) is another option . sort it out later */
Event *make_event(int type, char *name, int Nmax, int number_of_element, 
		  int *def, double *aS, double *aE, 
		  double deltaL, double deltaU, double pL, double pU)
{
    int i; 
    /* memory allocation */
    Event *event= (Event*) malloc(sizeof(Event));
    event->definition = (int*) malloc( number_of_element * sizeof(int));
    event->lowerbound = (int*) malloc( (Nmax+1) * sizeof(int));
    event->upperbound = (int*) malloc( (Nmax+1) * sizeof(int));
    if(!event->lowerbound || !event->upperbound || 
       !event->definition || !event) {
	printf("** Cannot allocate memory for make_event()\n.");
	return NULL;
    }
    event->type = type; /* 0= Adverse;  1= Efficacy */
    strcpy(event->name, name);

    for(i=0; i < number_of_element; i++)   
      event->definition[i]= (def == NULL) ? 0 : def[i];  
 
    event->number_of_element = number_of_element;
    event->Nmax = Nmax;
    event->deltaL = deltaL;  event->pL = pL;
    event->deltaU = deltaU;  event->pU = pU;

    /* compute the parameter of etaS ~ beta(aS, bS) and etaE ~ beta(aE, bE) */
    event->aS = 0.0;  event->bS = 0.0;
    event->aE = 0.0;  event->bE = 0.0;
    if(def != NULL) {
	for(i = 0; i < number_of_element; i++) {
	    if(def[i]==1) { 
		event->aS += aS[i]; 
		event->aE += aE[i]; 
	    }
	    if(def[i]==0) {
		event->bS += aS[i]; 
		event->bE += aE[i]; 
	    }
	}
    }
    for(i=0; i <= Nmax; i++)  {  
	event->lowerbound[i]=  -1;   /* the default boundary */
	event->upperbound[i]= i+1;   /* i = number of patients */
    }
    event->boundary_type = 4; /* default is 4 = unknown */
    event->child_count = 0;
    event->child = NULL;       /* parent-child link  */
    event->next = NULL;     /* sibling link. non-conditional  */
    event->parent = NULL;  /* first-level(unconditional) event */
    event->seed = NULL;   /* for remove event: (event|*)  */

    event->design = NULL;  /* will be done by add_event()   */
    return event;
}/*-------------------- End of make_event() -------------------------*/


void free_event(Event *event)
{
    if(event != NULL) {
	free_event(event->child); 
	free_event(event->next); /* maybe overdone , check it later */
	/* action */
	free(event->definition);   free(event->lowerbound); 
	free(event->upperbound);   free(event);
    }
}/*-------------------- End of free_event() -----------------------*/


/* make conditional event: (child | parent) */
/* key point: conditional event is itself an event */
/* this function automatically link the cond event, no need to add_event() */
void make_conditional_event(Event *new, Event *parent, int *flag)
{
    int i, k; char *temp; int number_of_element; 
    Event *cond_event;  Event *tail; 
    *flag = 0;
    k = check_conditional(new, parent);
    if (k < 3) { /* illegal */
	if (k==0) printf("** Null event in make_conditional_event()\n");
	if (k==1) printf("** Event '%s' does not intersect event '%s'\n",
			 new->name, parent->name);
	if (k==2) printf("** Identical event.\n");
	printf("** Conditional event '%s|%s' is not sensible.\n",
	       new->name, parent->name);
	*flag = FAIL;   return; /* bail out */
    } 
    else if (k >= 3) { /* legitimate, check futher if previous defined */
	Event *cond;
	for (cond = parent->child; cond != NULL; cond = cond->next) 
	  if (strcmp(cond->seed->name, new->name) == 0) {
	      printf("** Event '%s|%s' has already been defined.\n",
		     new->name, parent->name);
	      *flag = FAIL; return;
	  } 	
    }
    /* legal actions .......................................... */
    /* memory allocation */
    number_of_element = parent->design->number_of_element;
    cond_event = (Event*) malloc(sizeof(Event));
    cond_event->definition = (int*) malloc(number_of_element * sizeof(int));
    cond_event->lowerbound = (int*) malloc((parent->Nmax+1) * sizeof(int));
    cond_event->upperbound = (int*) malloc((parent->Nmax+1) * sizeof(int));
    if (!cond_event || !cond_event->definition || 
	!cond_event->lowerbound || !cond_event->upperbound) {
	printf("** Not enough memory for conditional event.\n");
	cond_event = NULL;
	*flag = FAIL;
	return;
    }
    cond_event->type = 2; /* 0 = Adverse;  1 = Efficacy  2 = unknown */
    /* naming */
    temp = cond_event->name;
    strcpy(temp, new->name);
    temp = strcat(temp, "|");
    temp = strcat(temp, parent->name);

    /* definition */  /* maybe bugs here, fix it later */
    /* need renormalization */
    for(i=0; i < number_of_element; i++)   
      cond_event->definition[i] = 
	new->definition[i] * parent->definition[i];
    
    /* need renormalization */
    /* smaller number_of_element ? */
    cond_event->number_of_element = parent->number_of_element;
    cond_event->Nmax = parent->Nmax;
    cond_event->Nmin = 1;  /* will be useful in cond monitoring */
    cond_event->deltaL = 0.0;  cond_event->pL = 0.0;
    cond_event->deltaU = 0.0;  cond_event->pU = 0.0;
    /* (aS, bS) can be computed */
    cond_event->aS = 0.0;     cond_event->bS = 0.0;
    cond_event->aE = 0.0;     cond_event->bE = 0.0;
    for(i=0; i < number_of_element; i++) {
	if(cond_event->definition[i] == 1) {
	    cond_event->aS += parent->design->aS[i]; 
	    cond_event->aE += parent->design->aE[i]; 
	}
	/* ask Peter if this is correct, later. I think (aS, bS) is
	   right. But (aE, bE) may need renormalization s.t. 
	   aE+bE = K (# of elements included in  'parent') */
	if(cond_event->definition[i] == 0 && parent->definition[i] == 1) {
	    cond_event->bS += parent->design->aS[i]; 
	    cond_event->bE += parent->design->aE[i]; 
	}
    }
    cond_event->boundary_type = 4;
    for(i=0; i <= cond_event->Nmax; i++)   {  
	cond_event->lowerbound[i]= -1;   /* the default boundary */
	cond_event->upperbound[i]=i+1;   /* i = number of patients */
    }
    /* legitimate child joins the family */ 
    /* a FILO list */
    /*    cond_event->next = parent->child;
	  if (parent->child != NULL)  parent->child->last = cond_event;
	  parent->child = cond_event;    
	  cond_event->last = NULL;  */
   
    /* a FIFO list */
    if (parent->child == NULL) {  /* becomes the first child */
	parent->child = cond_event;
	cond_event->next = NULL;
    }
    else { 
	for(tail = parent->child; tail->next != NULL; tail = tail->next)
	  ;
	tail->next = cond_event;
	cond_event->next = NULL;
    }

    cond_event->child = NULL;  
    cond_event->parent = parent;
    cond_event->seed = new;
    cond_event->design = parent->design;

    ++parent->child_count;
    ++cond_event->design->number_of_cond_event; /* # of cond events */
    parent->design->current_event = cond_event; /* maybe unnecessary */
    *flag = SUCCEED;
}/*-------------------- End of make_conditional_event() -------------*/


void remove_conditional_event(Event *cond_event)
{
    Event *temp;
    if (cond_event != NULL) {
	if (cond_event->parent->child == cond_event) { /* first child */
	    cond_event->parent->child = cond_event->next;
	}
	else {  /* not first child */
	    for(temp = cond_event->parent->child; temp->next != cond_event; 
		temp = temp->next)
	      ;
	    temp->next = cond_event->next;
	}
	--cond_event->parent->child_count;
	--cond_event->design->number_of_cond_event;
	free_event(cond_event); 
    }
}/*-------------------- End of remove_conditional_event() ------------*/

/* check if (child | parent) is legitimate:
   illegal:
   0 = null event(s);              
   1 = empty intersection;
   2 = 'child' == 'parent';

   legal :
   3 = 'child' is not a subset of 'parent', non-empty intersection;
   4 = OK;
*/

int check_conditional(Event *child, Event *parent)
{
    int i, j, sub, k = parent->number_of_element; 
    if (child == NULL || parent == NULL)  return 0;
    j = 0; sub = 1;
    for (i = 0; i < k; i++) {
	if ( parent->definition[i] < child->definition[i] ) 
	  sub = 0;  /*  'child' is not a subset of 'parent'; */
	if ( parent->definition[i] == child->definition[i] ) 
	  ++j;
    }
    if (j == k) return 2;  /* child == parent */
    if (j == 0) return 1;  /* empty intersection */
    /* OK */
    if (sub == 0) return 3;
    else return 4;
}/*------------------------ End of check_conditional() ---------------*/


