/* "utility.c" */
/* Time-stamp: <97/10/01 13:30:24 hgs> */
/* matrix operation and I/O facilities */

#include <stdio.h> 
#include <stdlib.h> 
#include <stddef.h>
#include <math.h> 
#include <string.h>
#include <assert.h>
#include "header.h"

/*-----------------------*
 * I/O facilities        *
 *-----------------------*/

/* *flag = 0 if fail. */
void print_design_in_file(Design *design, int *flag)
{
    FILE *fp;  char *filename;     Event *comp, *cond;

    filename = design->filename; 
    if(!(fp = fopen(filename, "w"))) {
        printf("\n** Can not write to file:\n\"%s\".\n"
	       "** Please check if the path is correct.\n", filename);
	*flag = 0;
	return;
    } else {
	printf("\n** The design parameters are stored in file:\n"
	       "** \"%s\"\n", filename);
	*flag = 1;
    }
    fprintf(fp, "\n(* DO NOT edit this file *)\n\n"
	    "Storage file for design: \"%s\"\n\n"
	    "Design Name: %s\n", filename, filename);
    /* ==================================  
       OUTPUT THE DATA INTO THE FILE    
       ==================================  */
    fprintf(fp, "Nmin = %d\nNmax = %d\n", design->Nmin, design->Nmax);
    fprintf(fp, "Number of elementary events = %d\n",
	    design->number_of_element);
    fprintf(fp, "Number of compound events = %d\n", design->number_of_event);
    fprintf(fp, "Number of conditional events = %d\n", 
	    design->number_of_cond_event);
    
    if(design->monitor_method == 1) {
	fprintf(fp, "\nMonitoring method = By cohort\n"
		"Cohort size = %d\n\n", design->cohort_size);
    }
    else if(design->monitor_method == 2) {
	  fprintf(fp, "\nMonitoring method = Periodic\n");
	  fprintf(fp, "Accrual rate = %.3lf\n", design->accrual_rate);
	  fprintf(fp, "Monitoring period = %.3lf\n",design->monitor_period);
	  fprintf(fp, "Response window = %.3f\n\n", design->response_window);
      }
      fprintf(fp, "Standard Prior Type: %s\n",	
	      (design->Sprior_type == 1)? "Single" : "Mixture"); 

      if (design->Sprior_type == 1) { /* single prior */
	  int i;
	  fprintf(fp, "Parameters of Elementary Events:\n"
		  "\n--------------------------------\n"
		  "       Prior Rx Parameters\n"
		  "     Standard   Experimental\n"
		  "    ----------  ------------\n");
	  for(i=0; i < design->number_of_element; i++) 
	    fprintf(fp, "a_S%d = %7.3lf\ta_E%d = %6.3lf\n", 
		    i+1, design->aS[i], i+1, design->aE[i]);
	  fprintf(fp, "--------------------------------\n");
      }
      else { /* mixture prior */
	  int i, j;
	  fprintf(fp, "Number of Mixture = %d\n\n", design->Sprior_type);
	  fprintf(fp, "Standard Prior Rx Parameters\t\tWeights\n"
		      "----------------------------\t\t--------\n");
	  for(i=0; i< design->Sprior_type; i++) {
	      fprintf(fp, "Dir_%d: ", i+1);
	      for(j=0; j< design->number_of_element; j++)
		fprintf(fp, "%3.3f\t", design->mix_aS[i][j]);
	      fprintf(fp, "\t%.3f\n", design->Sprior_weight[i]);
	  }
	  fprintf(fp, "\nExperimental Prior Rx Parameters\n"
		  "----------------------------------\n");
	  for(i=0; i < design->number_of_element; i++) 
	    fprintf(fp, "a_E%d = %3.3f\n", i+1, design->aE[i]); 
	  fprintf(fp, "\n"); 
      }
  /*    sumS = sumE = 0.0;
	for(i=0; i < design->number_of_element; i++) {
	sumS += design->aS[i];   sumE += design->aE[i];
	}
	for(i=0; i < design->number_of_element; i++) {
	if(design->Sprior_type == 1)
	fprintf(fp, "meanS%d = %.3lf\tmeanE%d = %.3lf\n", 
	i+1, design->aS[i]/sumS, i+1, design->aE[i]/sumE);
	fprintf(fp, "--------------------------------\n");
  */
      fprintf(fp, "\n-------------------\n"
	      "Compound Events:\n"
	      "-------------------\n");    
      for(comp = design->event0->next; comp != NULL; comp = comp->next) 
	print_event_in_file(comp, fp);
      fprintf(fp, "\n-------------------\n"
	      "Conditional Events:\n"
	      "-------------------\n");

      if (design->number_of_cond_event > 0) {
	  for(comp = design->event0->next; comp != NULL; comp = comp->next) 
	    for(cond = comp->child; cond != NULL; cond = cond->next)
	      print_event_in_file(cond, fp);

      } else  fprintf(fp, "None.\n");

      fprintf(fp, "=====================================================\n");
      (void)fclose(fp);
  }/*-------------------------------end of print_design_in_file()-------*/




  /* print out the design parameters */
  void print_design_on_screen(Design* design)
  {
      Event *comp, *cond;
      printf("\n==========================\n");
      printf("Review of Design\n");
      printf( "==========================\n\n");
      printf("Design Name:  %s\n", design->filename);
      printf("Nmin = %d\nNmax = %d\n", design->Nmin, design->Nmax);
      printf("Number of elementary events  = %d\n", design->number_of_element);
      printf("Number of compound events    = %d\n", design->number_of_event);
      printf("Number of conditional events = %d\n",design->number_of_cond_event);

      if(design->monitor_method == 1) {
	  printf("\nMonitoring method = By cohort\n");
	  printf("Cohort size = %d\n", design->cohort_size);
      }
      else if(design->monitor_method == 2) {
	  printf("\nMonitoring method = Periodic\n");
	  printf("Accrual rate = %.3lf\n", design->accrual_rate);
	  printf("Monitoring period = %.3lf\n", design->monitor_period);
	  printf("Response window = %.3f\n\n", design->response_window);
      }  
      printf("Standard Prior Type: %s\n",	
	     (design->Sprior_type == 1)? "Single" : "Mixture"); 

      if (design->Sprior_type == 1) { /* single prior */
	  int i;
	  printf("Parameters of Elementary Events:\n");
	  printf("\n--------------------------------\n");
	  printf("       Prior Rx Parameters\n");
	  printf("     Standard   Experimental\n");
	  printf("    ----------  ------------\n");
	  for(i=0; i < design->number_of_element; i++) 
	    printf("a_S%d = %7.3lf\ta_E%d = %6.3lf\n", 
	       i+1, design->aS[i], i+1, design->aE[i]);
	  printf("--------------------------------\n");
      }
      else {
	  int i, j;
	  printf("Number of Mixture = %d\n\n", design->Sprior_type);
	  printf("Standard Prior Rx Parameters\t\tWeights\n");
	  printf("----------------------------\t\t-------\n");
	  for(i=0; i< design->Sprior_type; i++) {
	      printf("Dir_%d: ", i+1);
	      for(j=0; j< design->number_of_element; j++)
		printf("%.3f\t", design->mix_aS[i][j]);
	      printf("\t%.3f\n", design->Sprior_weight[i]);
	  }
	  printf("\nExperimental Prior Rx Parameters\n");
	  printf("----------------------------------\n");
	  for(i=0; i < design->number_of_element; i++) 
	    printf("aE_%d = %6.3f\n", i+1, design->aE[i]); printf("\n"); 
      }
  /*   sumS = sumE = 0.0;
       for(i=0; i < design->number_of_element; i++) {
       sumS += design->aS[i];   sumE += design->aE[i];
       }
       for(i=0; i < design->number_of_element; i++)
       printf("meanS%d = %.3lf\tmeanE%d = %.3lf\n", 
       i+1, design->aS[i]/sumS, i+1, design->aE[i]/sumE);
       printf("--------------------------------\n");
  */
      printf("\n-------------------\n");
      printf("Compound Events:\n");
      printf("-------------------\n");  
      for(comp = design->event0->next; comp != NULL; comp = comp->next) 
	print_event_on_screen(comp);
      printf("\n-------------------\n");
      printf("Conditional Events:\n");
      printf("-------------------\n");
      if( design->number_of_cond_event > 0) {
	  for(comp = design->event0->next; comp != NULL; comp = comp->next) 
	    for(cond = comp->child; cond != NULL; cond = cond->next)
	      print_event_on_screen(cond);
      }
      else printf("None.\n");
      printf("=====================================================\n");
  } /*-------------- End of print_design_on_screen() ---------------------*/


  /* summary of basic setup */
  void print_event_on_screen(Event *event)
  {
      int i; double final_mean, mu;
      if(event == NULL) return;
      printf("\n--------------------\n");
      printf("Event Name: %s\n", event->name);
      printf("--------------------\n");

      printf("Elememtary Events Included: { ");
      for(i=0; i < event->number_of_element; i++) 
	if(event->definition[i] == 1) printf("A%d ",i+1);
      printf("}");
      if(event->parent != NULL) { 
	  printf(" | { ");
	  for(i=0; i < event->number_of_element; i++)
	    if(event->parent->definition[i] == 1) printf("A%d ", i+1);
	  printf("}");
      }

      printf("\nEvent Type: %s\n", (event->type==1)? "Efficacy": 
	     (event->type==0)? "Adverse": "Other");

      if (event->design->Sprior_type == 1) { /* single */
	  printf("a_S = %7.3lf, b_S = %7.3lf,\t", event->aS, event->bS);
	  printf("mu_S = a_S/(a_S + b_S) = %.3lf\n",
		 event->aS/(event->aS+event->bS));
	  printf("a_E = %7.3lf, b_E = %7.3lf,\t", event->aE, event->bE);
	  printf("mu_E = a_E/(a_E + b_E) = %.3lf\n",
		 event->aE/(event->aE+event->bE));
      }
      else { /* mixture */
	  int k = event->design->Sprior_type;
	  final_mean = 0.0;
    printf("\nDir()   weight   a_S       b_S         mu_S = a_S/(a_S + b_S)\n"
	     "------  ------  -------   -------     -----------------------\n");
	  for(i=0; i< k; i++){
	      printf("Dir_%d:\t", i+1);
	      printf("%.3f\t", event->design->Sprior_weight[i]);
	      printf("%7.3f\t%9.3f\t", event->mix_aS[i], event->mix_bS[i]); 
	      mu = event->mix_aS[i]/(event->mix_aS[i]+event->mix_bS[i]);
	      printf("\t%.3f\n", mu);
	      final_mean += event->design->Sprior_weight[i] * mu;
	  }
	  printf("Overall Prior Mean = w_1 * mu_{S,1} +...+ ");
	  printf("w_k * mu_{S,%d} = %.3f\n", k, final_mean); 
      }

      printf("\nBoundary Parameters:\n");
      if( event->boundary_type == 4 )
	printf("No Monitoring Boundary");
      else {
	  if(event->type == 1 ) { 
	      printf("Lower Efficacy Bound: delta_L = %.3lf  p_L = %.3lf\n",
		     event->deltaL, event->pL);
	      if(event->boundary_type == 1 || event->boundary_type == 4) 
		printf("Upper Efficacy Bound: None.\n");
	      else
		printf("Upper Efficacy Bound: delta_U = %.3lf  p_U = %.3lf\n",
		       event->deltaU, event->pU);
	  }
	  if(event->type == 0) 
	    printf("Upper Adverse Bound: delta_U = %.3lf  p_U = %.3lf\n",
		   event->deltaU, event->pU);
      }	    
      printf("\n");
  }/*----- End of print_event_on_screen() -----------------------*/ 


  void print_event_in_file(Event *event, FILE *fp)
  {
      int i; double final_mean, mu;

      if(event == NULL || fp == NULL) return;
      fprintf(fp, "\n--------------------\n");
      fprintf(fp, "Event Name: %s\n", event->name);
      fprintf(fp, "--------------------\n");
      fprintf(fp, "Elememtary Events Included: { ");
      for(i=0; i < event->number_of_element; i++) 
	if(event->definition[i] == 1) fprintf(fp, "A%d ",i+1);
      fprintf(fp, "}");
      if(event->parent != NULL) {
	  fprintf(fp, " | { ");
	  for(i=0; i < event->number_of_element; i++)
	    if(event->parent->definition[i] == 1) fprintf(fp, "A%d ", i+1);
	  fprintf(fp, "}");
      }
      fprintf(fp, "\nDefinition code: ");
      for(i=0; i < event->number_of_element; i++) 
	fprintf(fp, "%d", event->definition[i]);

      if (event->parent != NULL) {
	  fprintf(fp, "\nParent name: %s", event->parent->name); 
	  fprintf(fp, "\nParent code: ");
	  for(i=0; i < event->number_of_element; i++) 
	    fprintf(fp, "%d", event->parent->definition[i]);

	  fprintf(fp, "\nSeed code: ");
	  for(i=0; i < event->number_of_element; i++) 
	    fprintf(fp, "%d", event->seed->definition[i]);
      }
      fprintf(fp, "\nEvent Type: %s\n", (event->type==1)? "Efficacy": 
	     (event->type==0)? "Adverse": "unknown");
      if(event->design->Sprior_type == 1) { /* single prior */   
	  fprintf(fp, "a_S = %7.3lf\tb_S = %7.3lf\t", event->aS, event->bS);
	  fprintf(fp, "mu_S = a_S/(a_S + b_S) = %.3lf\n",
		  event->aS/(event->aS+event->bS));
	  fprintf(fp, "a_E = %7.3lf\tb_E = %7.3lf\t", event->aE, event->bE);
	  fprintf(fp, "mu_E = a_E/(a_E + b_E) = %.3lf\n",
		  event->aE/(event->aE+event->bE));
      }
      else { /* mixture */
	  final_mean = 0.0; 
   fprintf(fp,"\nDir()   weight   a_S       b_S         mu_S = a_S/(a_S + b_S)\n"
	 "------  ------  -------   -------     -----------------------\n");
	  for(i=0; i< event->design->Sprior_type; i++){
	      fprintf(fp, "Dir_%d:\t", i+1);
	      fprintf(fp,"%.3f\t", event->design->Sprior_weight[i]);
	      fprintf(fp,"%7.3f\t%9.3f\t", event->mix_aS[i], event->mix_bS[i]); 
	      mu = event->mix_aS[i]/(event->mix_aS[i]+event->mix_bS[i]);
	      fprintf(fp,"\t%.3f\n", mu);
	      final_mean += event->design->Sprior_weight[i] * mu;
	  }
	  fprintf(fp, "Overall Prior Mean = w_1 * mu_{S,1} +...+ ");
	  fprintf(fp, "w_k * mu_{S,%d} = %.3f\n", 
		  event->design->Sprior_type, final_mean); 
      }
      fprintf(fp, "\nBoundary Parameters:\n");
      fprintf(fp, "boundary type = %d\n", event->boundary_type);
      if(event->boundary_type == 1 || event->boundary_type == 2) {
	  fprintf(fp, "delta_L = %.3lf\t\tp_L = %.3lf\n", 
		  event->deltaL, event->pL);
      }
      if(event->boundary_type == 2 || event->boundary_type == 3) {
	  fprintf(fp, "delta_U = %.3lf\t\tp_U = %.3lf\n", 
		  event->deltaU, event->pU);
      }
      fprintf(fp, "\n");
  }/*--------------------------- End of print_event_in_file()------------*/ 


  /* print out the boundary for each event in the design to a file */
  void print_boundary_in_file(Design *design)
  {
      FILE *fp;  Event *comp, *cond;
      char buf[100],  *one;  

      sprintf(buf, "%s", design->filename);
      one = buf;
      one = strcat(one, ".bnds");
      printf("\n** The boundaries are stored in file:\n** \"%s\"\n", one);
      fp = fopen(one, "w");
      if(!fp) {
	  printf("\n** Could not open file \n'%s' for "
		 "storing the boundaries!!\n", one);
	  return;
      }            
      fprintf(fp, "\n--------------------------------------------\n"
	      	     "Boundaries of design:\n\"%s\"\n", design->filename);
     fprintf(fp, "--------------------------------------------\n" 
	     "Basic setup:\n-------------\n"
	     "N.min = %d\nN.max = %d\n", design->Nmin, design->Nmax);
     if(design->monitor_method == 1)
       fprintf(fp, "cohort size = %d\n", design->cohort_size);
     else if(design->monitor_method == 2) 
       fprintf(fp, "accrual rate = %.3f\nmonitoring period = %.3f\n",
	       design->accrual_rate, design->monitor_period);

     for(comp = design->event0->next; comp != NULL; comp = comp->next) 
       print_event_boundary_in_file(comp, fp);

     for(comp = design->event0->next; comp != NULL; comp = comp->next) 
       for(cond = comp->child; cond != NULL; cond = cond->next) 
	 print_event_boundary_in_file(cond, fp);

     (void)fclose(fp);
 } /*------------------ End of print_boundary_in_file() -----------------*/

 void print_boundary_on_screen(Design *design)
 {
     Event *comp, *cond;

     printf("\n--------------------------------------------\n"
	    "Boundaries of design:\n\"%s\"\n", design->filename);
     printf("--------------------------------------------\n"
	    "Basic setup:\n"
	    "-------------\n"
	    "N.min = %d\nN.max = %d\n", design->Nmin, design->Nmax);
     if(design->monitor_method == 1)
       printf("cohort size = %d\n", design->cohort_size);
     else if(design->monitor_method == 2) 
       printf("accrual rate = %.3f\nmonitoring period = %.3f\n",
	       design->accrual_rate, design->monitor_period);

     for(comp = design->event0->next; comp != NULL; comp = comp->next) 
       print_event_boundary_on_screen(comp);

     for(comp = design->event0->next; comp != NULL; comp = comp->next) 
       for(cond = comp->child; cond != NULL; cond = cond->next) 
	 print_event_boundary_on_screen(cond);

 } /*-------------------------- End of print_boundary_on_screen() ---------*/


 void print_event_boundary_in_file(Event *event, FILE *fp)
 {
     int j, nmax;
     nmax = event->design->Nmax;
     if(fp == NULL) { 
	 printf("cannot open file for boundary printing.\n");
	 return;
     }
     fprintf(fp, "\nThe boundaries for : %s\n", event->name);
     fprintf(fp, "------------------------------\n"
	     "Event type : %s\n", (event->type == 1)? "Efficacy" : 
	     ((event->type == 0)? "Adverse" : "unknown"));

     if(event->boundary_type == 3 || event->boundary_type == 4
	|| event->boundary_type == 0 )
       fprintf(fp, "Lower bound: None\n");
     else
       fprintf(fp, "Lower bound: delta_L = %.3lf, p_L = %.3lf\n",
	       event->deltaL, event->pL);
     if(event->boundary_type == 1 || event->boundary_type == 4
       || event->boundary_type == 0 )
      fprintf(fp, "Upper bound: None\n");
     else
       fprintf(fp, "Upper bound: delta_U = %.3lf, p_U = %.3lf\n",
	       event->deltaU, event->pU);
     fprintf(fp, "\nStopping bounds\n------------------\n");
     if(event->boundary_type == 1 || event->boundary_type == 2) {
	 if ( event->parent == NULL ) /* compound event */
	   fprintf(fp, "Stop the trial if "
		   "[# patients with %s]/[# patients scored]\n"
		   "is less than or equal to the lower bound.\n", event->name);
	 else  /* conditional event */
	   fprintf(fp, "Stop the trial if "
		   "[# patients with %s]/[# patients with %s]\n"
		   "is less than or equal to the lower bound.\n",
		   event->seed->name, event->parent->name);
     }
     if(event->boundary_type == 3 || event->boundary_type == 2) { 
	 if ( event->parent == NULL ) /* compound event */
	   fprintf(fp,"Stop the trial if "
		   "[# patients with %s]/[# patients scored]\n"
		   "is greater than or equal to the upper bound.\n",
		   event->name);
	 else /* conditional event */
	   fprintf(fp, "Stop the trial if "
		   "[# patients with %s]/[# patients with %s]\n"
		   "is greater than or equal to the upper bound.\n",
		   event->seed->name, event->parent->name);
     }
     if(event->boundary_type == 1) { /* lowerbounds only */
	 fprintf(fp, "\n-----------------------------------------------\n");
	 if (event->parent == NULL) /* compound event */
	   fprintf(fp, "  Lower bound      Number of patients\n");
	 else /* conditional */
	   fprintf(fp, "  Lower bound      Number of patients with %s\n",
		   event->parent->name);
	 fprintf(fp, " --------------   ----------------------------\n");
	 if ( event->lowerbound[event->design->Nmax - 1] < 0 )
	   fprintf(fp, "No lower bound for current setting.\n");

	 for (j = 1; j < event->design->Nmax; j++) 
	   if ( j % event->design->cohort_size == 0 &&  
	      event->lowerbound[j] >= 0) {
	       fprintf(fp,"      %3d          %3d\n", event->lowerbound[j], j);
	   }
     }
     else if (event->boundary_type == 2) { /* both */
	 fprintf(fp, "\n-----------------------------------------------\n");
	 if (event->parent == NULL) /* compound event */
	   fprintf(fp, "  Lower bound   Upper bound   Number of patients\n");
	 else
	   fprintf(fp, "  Lower bound   Upper bound   Number of patients with %s\n",
		   event->parent->name);
	 fprintf(fp, " -------------  ------------  --------------------\n");
	 if (event->upperbound[nmax-1] > nmax-1 && 
	    event->lowerbound[nmax-1] < 0 )
	   fprintf(fp, "No boundary for current setting.\n");

	 for(j = 1; j < event->design->Nmax; j++) 
	   if( j % event->design->cohort_size == 0 ) {
	       if(event->lowerbound[j] >= 0 && event->upperbound[j] > j)
		 fprintf(fp, "      %3d          ---           %3d\n",
			event->lowerbound[j], j); 
	       else if(event->lowerbound[j] >= 0 && event->upperbound[j] <= j)  
		 fprintf(fp, "      %3d          %3d           %3d\n",
			event->lowerbound[j], event->upperbound[j], j);
	       else if(event->lowerbound[j] < 0 && event->upperbound[j] <= j)
		 fprintf(fp, "      ---          %3d           %3d\n",
			event->upperbound[j], j);
	   }	
     }
     else if (event->boundary_type == 3) { /* upper only */
	 fprintf(fp, "\n-----------------------------------------------\n");
	 if (event->parent == NULL) /* compound event */
	   fprintf(fp, "  Upper bound      Number of patients\n");
	 else /* conditional */
	   fprintf(fp, "  Upper bound      Number of patients with %s\n",
		   event->parent->name);
	 fprintf(fp, " --------------   ----------------------------\n");

	 if(event->upperbound[nmax-1] > nmax-1)
	   fprintf(fp, "No upper bound for current setting.\n");

	 for(j = 1; j < event->design->Nmax; j++) { 
	     if( j % event->design->cohort_size == 0 &&
		/* event->upperbound[j]!=event->upperbound[j-1] && */
		event->upperbound[j] <= j) {
		 fprintf(fp,"      %3d          %3d\n",event->upperbound[j], j);
	     }
	 }
     }
     fprintf(fp, "-----------------------------------------------\n\n");
 } /*------------------ End of print_event_boundary_in_file() --------------*/


 void print_event_boundary_on_screen(Event *event)
 {
     int j, nmax;
     nmax = event->design->Nmax;
     printf("\nThe boundaries for : %s\n", event->name);
     printf("------------------------------\n"
	    "Event type : %s\n", (event->type == 1)? "Efficacy" : 
	    ((event->type == 0)? "Adverse" : "unknown"));

     if(event->boundary_type == 3 || event->boundary_type == 4
	|| event->boundary_type == 0 )
       printf("Lower bound: None\n");
     else
       printf("Lower bound: delta_L = %.3lf, p_L = %.3lf\n",
	       event->deltaL, event->pL);
     if(event->boundary_type == 1 || event->boundary_type == 4
	|| event->boundary_type == 0 )
       printf("Upper bound: None\n");
     else
       printf("Upper bound: delta_U = %.3lf, p_U = %.3lf\n",
	       event->deltaU, event->pU);
     printf("\nStopping bounds\n------------------\n");
     if(event->boundary_type == 1 || event->boundary_type == 2) {
	 if ( event->parent == NULL ) /* compound event */
	   printf("Stop the trial if [# patients with %s]/[# patients scored]\n"
		  "is less than or equal to the lower bound.\n", event->name);
	 else  /* conditional event */
	   printf("Stop the trial if "
		  "[# patients with %s]/[# patients with %s]\n"
		  "is less than or equal to the lower bound.\n",
		  event->seed->name, event->parent->name);
     }
     if(event->boundary_type == 3 || event->boundary_type == 2) { 
	 if ( event->parent == NULL ) /* compound event */
	   printf("Stop the trial if [# patients with %s]/[# patients scored]\n"
		  "is greater than or equal to the upper bound.\n", 
		  event->name);
	 else  /* conditional event */
	   printf("Stop the trial if "
		  "[# patients with %s]/[# patients with %s]\n"
		  "is greater than or equal to the upper bound.\n",
		  event->seed->name, event->parent->name);
     }	
     if(event->boundary_type == 1) { /* lower only */
	 if (event->parent == NULL) /* compound event */
	   printf( "  Lower bound      Number of patients\n");
	 else /* conditional */
	   printf( "  Lower bound      Number of patients with %s\n",
		   event->parent->name);
	 printf( " --------------   ----------------------------\n");
	 if( event->lowerbound[nmax - 1] < 0 )
	   printf("No lower bound for current setting.\n");

	 for(j = 1; j < event->design->Nmax; j++) 
	   if( j % event->design->cohort_size == 0 &&
	      /* event->lowerbound[j]!=event->lowerbound[j-1] && */
	      event->lowerbound[j] >= 0)
	     printf("      %3d          %3d\n", event->lowerbound[j], j);
     }  
     else if (event->boundary_type == 2) { /* both */
	 printf( "\n-----------------------------------------------\n");
	 if (event->parent == NULL) /* compound event */
	   printf( "  Lower bound   Upper bound   Number of patients\n");
	 else
	   printf( "  Lower bound   Upper bound   Number of patients with %s\n",
		   event->parent->name);
	 printf( " -------------  ------------  --------------------\n");
	 if(event->upperbound[nmax-1] > nmax-1 && 
	    event->lowerbound[nmax-1] < 0 )
	   printf("No boundary for current setting.\n");

	 for(j = 1; j < nmax; j++) 
	   if( j % event->design->cohort_size == 0 ) {
	       if(event->lowerbound[j] >= 0 && event->upperbound[j] > j)
		 printf("      %3d          ---           %3d\n",
			event->lowerbound[j], j); 
	       else if(event->lowerbound[j] >= 0 && event->upperbound[j] <= j)  
		 printf("      %3d          %3d           %3d\n",
			event->lowerbound[j], event->upperbound[j], j);
	      else if(event->lowerbound[j] < 0 && event->upperbound[j] <= j)
		 printf("      ---          %3d           %3d\n",
			event->upperbound[j], j);
	   }
     }
     else if (event->boundary_type == 3) { /* upper only */
	 printf( "\n-----------------------------------------------\n");
	 if (event->parent == NULL) /* compound event */
	   printf( "  Upper bound      Number of patients\n");
	 else /* conditional */
	   printf( "  Upper bound      Number of patients with %s\n",
		   event->parent->name);
	 printf( " --------------   ----------------------------\n");
	 if(event->upperbound[nmax-1] > nmax-1)
	     printf("No upper bound for current setting.\n");

	 for(j = 1; j < event->design->Nmax; j++) 
	   if( j % event->design->cohort_size == 0  &&
	      event->upperbound[j] <= j 
	      /* && event->upperbound[j]!=event->upperbound[j-1] */)
	     printf("    %3d         %3d\n", event->upperbound[j], j);
     }    
     printf("-----------------------------------------------\n\n");
 }/*------------------ End of print_event_boundary_on_screen()-----------*/


 /*------------------------------------------------*
  *   matrix operation:                            *
  *------------------------------------------------*/
 int **imatrix(int nrow, int ncol)
 {
     int i, j, **matrix, *body;
     matrix =  (int**) malloc(nrow * sizeof(int*));
     if(!matrix) {
	 printf("** Warning: not enough memory for imatrix()\n");
	 return NULL;
     }
     body = (int*) malloc(nrow * ncol * sizeof(int)); 
     if(!body) { 
	 printf("** Warning: not enough memory for imatrix()\n");
	 free(matrix);
	 return NULL;
     }
     matrix[0] = body;
     for (i = 1; i < nrow; i++)  matrix[i] = matrix[i-1] + ncol;
     for (i = 0; i < nrow; i++) for (j = 0; j < ncol; j++)
       matrix[i][j] = 0;
     return matrix;
 }

 double **dmatrix(int nrow, int ncol)
 {
     int i, j; double **matrix, *body;
     matrix = (double**) malloc((size_t)(nrow * sizeof(double*)));
     if(!matrix) { 
	 printf("** Warning: not enough memory for dmatrix()\n"); 
	 return NULL; 
     }
     body = (double*) malloc(nrow * ncol * sizeof(double));
     if(!body) {
	 printf("** Warning: not enough memory for dmatrix()\n");
	 free(matrix);
	 return NULL; /* NULL */
     }
     matrix[0] = body;
     for (i = 1; i < nrow; i++) 
       matrix[i] = matrix[i-1] + ncol;
     for (i = 0; i < nrow; i++) for (j = 0; j < ncol; j++)
       matrix[i][j] = 0e0;
     return matrix;
 }

 void free_matrix(void **matrix)
 {
     free((void*) matrix[0]);
     free((void*) matrix);
 }

 void print_imatrix(int **matrix, int nrow, int ncol)
 {
     int i, j;
     for(j=0; j<ncol; j++) printf("\t[%d]", j);
     for(i=0; i<nrow; i++) { 
	 printf("\n[%d]", i); 
	 for(j=0; j<4; j++)  printf("\t%2d", matrix[i][j]); 
     } 
     printf("\n"); 
 }
 void print_dmatrix(double **matrix, int nrow, int ncol)
 { 
     int i, j;
     for(j=0; j<ncol; j++) printf("\t[%d]", j);
     for(i=0; i<nrow; i++) { 
	 printf("\n[%d]", i); 
	 for(j=0; j<4; j++)  printf("\t%.3f", matrix[i][j]); 
     } 
     printf("\n"); 
 }




 /* Input facilities */

 /*  get_an_integer is a stupid input routine which prompts the user
     for input, and allows only integer values to be input.       
     return -2 if choose "previous", otherwise return the integer */ 
 int get_an_integer(void)
 {
     char buf[100]; 
     while(1) {
	 if (scanf("%s",buf) == 0) { /* no input from stdin */
	     printf("This is an invalid value, please re-enter\n");
	     continue;
	 }
	 if ((strcmp(buf,"exit")==0)||(strcmp(buf,"quit")==0)) {
	     char buf1[20];
	     printf("Are you sure? (plesase type 'yes' or 'no') > ");
	     (void) scanf("%s",buf1);
	     if(strcmp(buf1,"yes") == 0) exit(0);
	     else return -99;
	 } 
	 else if(strcmp(buf, "back") == 0) return -2;
	 /* else if(strcmp(buf, "-2") == 0)  continue; */
	else if(strcmp(buf, "y") == 0 || strcmp(buf,"yes")==0) return YES;
	else if(strcmp(buf, "n") == 0 || strcmp(buf,"no") ==0) return NO;
	else return(atoi(buf));
    } 
}/*-------- End of get_an_integer() --------------*/
 

/* get_a_double prompts the user for input */
double get_a_double(double min, double max)
{
    double value;    char buf[100]; 

    if (max - min < 1e-10) return min;
    if (scanf("%s",buf)==0) {  /* no input from stdin */
	printf("Invalid input...please try again\n");
	return get_a_double(min, max);
    }   
    if ( strcmp(buf, "back") == 0 )  return -2e0; 
    if ( strcmp(buf, "-1") == 0 || strcmp(buf, "end") == 0 ) 
      return -1e0;  /* break point */
    if (strcmp(buf,"exit") == 0 || strcmp(buf, "quit") == 0) {
	char buf1[20];
	printf("Are you sure? (please type 'yes' or 'no')\n");
	(void) scanf("%s",buf1);
	if (strcmp(buf1,"yes") == 0)  exit(0); 
	else return get_a_double(min, max);
    } 
    value = strtod(buf, NULL);        
    if (value < min - 1e-7 || value > max + 1e-7) {
	printf("\nplease enter a number between %.3lf and %.3lf. > ",
	       min, max);
	return get_a_double(min, max);
    }   
    printf("\n");
    return value;
}/*----------------- End of get_a_double() ---------------------------*/

/* read the file "filename" and put in char *text */
/* bugs: if "filename" is a directory, the function will break down */
char *get_file(char *filename)
{  
    FILE *fp;  int ltext;   char *text;

    if (!(fp=fopen(filename,"r")))  {
	printf("\n** Cannot open file:\n\"%s\"\n", filename);
	return NULL;		
    } 
    else {
	/* count the number of characters in the file */
	ltext = 0;
	while (fgetc(fp) != EOF) ++ltext;
	if (ltext == 0) return NULL;
	/* allocate memory for the variables text */
	text = (char*) malloc( (ltext+1) * sizeof(char));

	if ( text == NULL ) {
	    perror("Cannot alloc enough space for get_file()\n");
	    (void)fclose(fp);
	    return NULL;
	}
	
	/* read in the characters from the file  */
	rewind(fp);
	if ((int)fread(text, sizeof(char), (size_t)ltext, fp) == 0)
	  perror("Warning by get_file(): may not have read entire file!\n");
	
	text[ltext] = '\0'; /* the end */
	
	(void)fclose(fp);
	return text;
    }
}/*------------------------------------------end of get_file()-------*/


/* return "token = %f"  */
/* char *get_data(char *text, char *token) */
int get_data(char *text, char *token, char *buf)
{
    char *location, *end, *term;
    int i, token_length, value_length;
    /*    char *buf = (char*) malloc( 200 * sizeof(char)); 
	  if (!buf) return buf; */
    /*   sprintf(buf,"");   */
    term = "\0";
    token_length = strlen(token)+1;
    location = strstr(text, token);
   /* if (!(location = strstr(text, token)) ) { */
    if (location == NULL) {
	printf("** Cannot find \"%s\".\n", token); 
	return FAIL;
    }  
    location += token_length;
     end = strstr(location,"\n");
    value_length = (int)(strlen(location) - strlen(end));
    
    for (i=0;i<value_length;i++) buf[i] = *(location+i);
    
    buf[value_length] = *term;
    return SUCCEED;
    /* return buf; */
}/*----------------------------------------end of get_data()----*/


/* move the current_event pointer to the 
   event indexed as index = 1, 2,..., tot */
int get_event(int index, Design *design)
{
    int tot, count; Event *comp, *cond;
    tot = design->number_of_event + design->number_of_cond_event;

    if (index < 0 || index > tot) return 0;
    count = 0;
    /* compound events */
    for(comp = design->event0->next; comp != NULL; comp = comp->next) {
	++count;
	if (index == count) {
	    design->current_event = comp;
	    return 1;
	}
    }  
    /* conditional events */
    for(comp = design->event0->next; comp != NULL; comp = comp->next) 
      for(cond = comp->child; cond != NULL; cond = cond->next)	{ 
	  ++count;
	  if (index == count) {
	      design->current_event = cond;
	      return 1;
	  }  
      }  
    return 0;
}/*---------------------------------------------end of get_event()------*/

/* get_event() + interface *//* this function does the same thing as
   traverse_event_list() in structure.c */
void choose_event(Design *design)
{
    int index, option, option2; Event *comp, *cond;
    index = 1;
    printf("\n");
    for(comp = design->event0->next; comp != NULL; comp = comp->next) {
	printf("(%d) %s\n", index, comp->name);
	++index;
    }
    if(design->number_of_cond_event != 0) {
	for(comp = design->event0->next; comp != NULL; comp = comp->next) 
	  for(cond = comp->child; cond != NULL; cond = cond->next) {
	      printf("(%d) %s\n", index, cond->name);
	      ++index;
	  }
    }
    printf("> ");
    option = get_an_integer();
    if(option >=1 && option <= index - 1) {
	if( get_event(option, design) == SUCCEED ) {
	    printf("You choose event: \"%s\"\n", design->current_event->name);
	    printf("Confirmed? (Yes = 1; No = 0) > ");
	    option2 = get_an_integer();
	    if(option2 == 0) choose_event(design);
	}
    } else { 
	printf("Please type an integer between 1 and %d.\n", index-1);
	choose_event(design);
    }
}/*---------------------------------------------end of choose_event()---*/

void print_event_list(Design *design)
{
    Event *comp, *cond; int j, tot;
    tot = design->number_of_event + design->number_of_cond_event;
    printf("\n---------------------------------------\n"
	   "List of events for design:\n\"%s\"\n"
	   "-----------------------------------------\n",design->filename);
    if (tot == 0) {
	printf("No event defined for design:\n\"%s\"\n", design->filename); 
	return; 
    }
    printf("Compound events:\n"
	   "----------------\n");	 
    j = 0; 
    for(comp = design->event0->next; comp != NULL; comp = comp->next)
      printf("(%d) %s\n", ++j, comp->name);

    assert(j == design->number_of_event);
    if (design->number_of_cond_event != 0) {
	printf("Conditional events:\n"
	       "-------------------\n");	
	for(comp = design->event0->next; comp != NULL; comp = comp->next)
	  for(cond = comp->child; cond != NULL; cond = cond->next) 
	    printf("(%d) %s\n", ++j, cond->name);
    }
    assert(j == tot);
    printf("======================\n");
}
