/* file "zrorj.c" */
/* Time-stamp: <97/05/14 17:58:37 hgs> */
/* FORTRAN subroutine that find the zeros of a function */

#include <stdio.h> 
#include <stdlib.h> 
#include <stddef.h>
#include <math.h> 
#include <string.h>
#include <assert.h>

#include "ranlib.h"
 
#define LPROTOTYPE 
#define fNULL 0
#define fmax(a,b) ((a) >= (b) ? (a) : (b))
#define ftnabs(x) ((x) >= 0 ? (x) : -(x))
/* extern float fsign( *//* float num, float sign */  /* ); */


/*
**********************************************************************
 
     SUBROUTINE ZROR(STATUS, X, FX, XLO, XHI, QLEFT, QHI)
          ZeRo of a function -- Reverse Communication
 
 
                              Function
 
 
     Performs the zero finding.  STZROR must have been called before
     this routine in order to set its parameters.
 
 
                              Arguments
 
 
     STATUS <--> At the beginning of a zero finding problem, STATUS
                 should be set to 0 and ZROR invoked.  (The value
                 of other parameters will be ignored on this call.)
 
                 When ZROR needs the function evaluated, it will set
                 STATUS to 1 and return.  The value of the function
                 should be set in FX and ZROR again called without
                 changing any of its other parameters.
 
                 When ZROR has finished without error, it will return
                 with STATUS 0.  In that case (XLO,XHI) bound the answe
 
                 If ZROR finds an error (which implies that F(XLO)-Y an
                 F(XHI)-Y have the same sign, it returns STATUS -1.  In
                 this case, XLO and XHI are undefined.
                         INTEGER STATUS
 
     X <-- The value of X at which F(X) is to be evaluated.
                         REAL X
 
     FX --> The value of F(X) calculated when ZROR returns with
            STATUS = 1.
                         REAL FX
 
     XLO <-- When ZROR returns with STATUS = 0, XLO bounds the
             inverval in X containing the solution below.
                         REAL XLO
 
     XHI <-- When ZROR returns with STATUS = 0, XHI bounds the
             inverval in X containing the solution above.
                         REAL XHI
 
     QLEFT <-- .TRUE. if the stepping search terminated unsucessfully
                at XLO.  If it is .FALSE. the search terminated
                unsucessfully at XHI.
                    QLEFT is LOGICAL
 
     QHI <-- .TRUE. if F(X) .GT. Y at the termination of the
              search and .FALSE. if F(X) .LT. Y at the
              termination of the search.
                    QHI is LOGICAL
 
**********************************************************************
     .. Scalar Arguments ..
*/
static void E0000(IENTRY,status,x,fx,xlo,xhi,qleft,qhi,zabstl,zreltl,zxhi,zxlo)
int IENTRY;
int *status;
float *x,*fx,*xlo,*xhi,*zabstl,*zreltl,*zxhi,*zxlo;
unsigned long *qleft,*qhi;
{
/*
     ..
     .. Save statement ..
     ..
     .. Local Scalars ..
     ..
     .. Intrinsic Functions ..
     ..
     .. Statement Functions ..
*/
#define ftol(zx) (0.5 * fmax(abstol,reltol * ftnabs((zx))))
static float a,abstol,b,c,d,fa,fb,fc,fd,fda,fdb,m,mb,p,q,reltol,tol,w,xxhi,xxlo;
static int ext,i99999;
static unsigned long first,qrzero;
    switch(IENTRY){case 0: goto ZROR; case 1: goto STZROR;}
ZROR:
/*
     ..
     .. Executable Statements ..
*/
    if(*status > 0) goto S280;
    *xlo = xxlo;
    *xhi = xxhi;
    b = *x = *xlo;
/*
     GET-FUNCTION-VALUE
*/
    i99999 = 1;
    goto S270;
S10:
    fb = *fx;
    *xlo = *xhi;
    a = *x = *xlo;
/*
     GET-FUNCTION-VALUE
*/
    i99999 = 2;
    goto S270;
S20:
/*
     Check that F(ZXLO) < 0 < F(ZXHI)  or
                F(ZXLO) > 0 > F(ZXHI)
*/
    if(!(fb < 0.0)) goto S40;
    if(!(*fx < 0.0)) goto S30;
    *status = -1;
    *qleft = *fx < fb;
    *qhi = 0;
    return;
S40:
S30:
    if(!(fb > 0.0)) goto S60;
    if(!(*fx > 0.0)) goto S50;
    *status = -1;
    *qleft = *fx > fb;
    *qhi = 1;
    return;
S60:
S50:
    fa = *fx;
    first = 1;
S70:
    c = a;
    fc = fa;
    ext = 0;
S80:
    if(!(ftnabs(fc) < ftnabs(fb))) goto S100;
    if(!(c != a)) goto S90;
    d = a;
    fd = fa;
S90:
    a = b;
    fa = fb;
    *xlo = c;
    b = *xlo;
    fb = fc;
    c = a;
    fc = fa;
S100:
    tol = ftol(*xlo);
    m = (c + b)*.5;
    mb = m - b;
    if(!(ftnabs(mb) > tol)) goto S240;
    if(!(ext > 3)) goto S110;
    w = mb;
    goto S190;
S110:
    tol = fsign(tol,mb);
    p = (b - a) * fb;
    if(!first) goto S120;
    q = fa - fb;
    first = 0;
    goto S130;
S120:
    fdb = (fd - fb) / (d - b);
    fda = (fd - fa) / (d - a);
    p = fda * p;
    q = fdb * fa - fda * fb;
S130:
    if(!(p < 0.0)) goto S140;
    p = -p;
    q = -q;
S140:
    if(ext == 3) p *= 2.0;
    if(!(p * 1.0 == 0.0 || p <= q * tol)) goto S150;
    w = tol;
    goto S180;
S150:
    if(!(p < mb * q)) goto S160;
    w = p / q;
    goto S170;
S160:
    w = mb;
S190:
S180:
S170:
    d = a;
    fd = fa;
    a = b;
    fa = fb;
    b += w;
    *xlo = b;
    *x = *xlo;
/*
     GET-FUNCTION-VALUE
*/
    i99999 = 3;
    goto S270;
S200:
    fb = *fx;
    if(!(fc * fb >= 0.0)) goto S210;
    goto S70;
S210:
    if(!(w == mb)) goto S220;
    ext = 0;
    goto S230;
S220:
    ext += 1;
S230:
    goto S80;
S240:
    *xhi = c;
    qrzero = fc >= 0.0 && fb <= 0.0 || fc < 0.0 && fb >= 0.0;
    if(!qrzero) goto S250;
    *status = 0;
    goto S260;
S250:
    *status = -1;
S260:
    return;
STZROR:
/*
**********************************************************************
     SUBROUTINE STZROR( XLO, XHI, ABSTOL, RELTOL )
          SeT ZeRO finder - Reverse communication version
                              Function
     Sets quantities needed by ZROR.  The function of ZROR
     and the quantities set is given here.
     Concise Description - Given a function F
     find XLO such that F(XLO) = 0.
          More Precise Description -
     Input condition. F is a real function of a single
     real argument and XLO and XHI are such that
          F(XLO)*F(XHI)  .LE.  0.0
     If the input condition is met, QRZERO returns .TRUE.
     and output values of XLO and XHI satisfy the following
          F(XLO)*F(XHI)  .LE. 0.
          ABS(F(XLO)  .LE. ABS(F(XHI)
          ABS(XLO-XHI)  .LE. TOL(X)
     where
          TOL(X) = MAX(ABSTOL,RELTOL*ABS(X))
     If this algorithm does not find XLO and XHI satisfying
     these conditions then QRZERO returns .FALSE.  This
     implies that the input condition was not met.
                              Arguments
     XLO --> The left endpoint of the interval to be
           searched for a solution.
                    XLO is REAL
     XHI --> The right endpoint of the interval to be
           for a solution.
                    XHI is REAL
     ABSTOL, RELTOL --> Two numbers that determine the accuracy
                      of the solution.  See function for a
                      precise definition.
                    ABSTOL is REAL
                    RELTOL is REAL
                              Method
     Algorithm R of the paper 'Two Efficient Algorithms with
     Guaranteed Convergence for Finding a Zero of a Function'
     by J. C. P. Bus and T. J. Dekker in ACM Transactions on
     Mathematical Software, Volume 1, no. 4 page 330
     (Dec. '75) is employed to find the zero of F(X)-Y.
**********************************************************************
*/

    xxlo = *zxlo;
    xxhi = *zxhi;
    abstol = *zabstl;
    reltol = *zreltl;

    return;
S270:
/*
     TO GET-FUNCTION-VALUE
*/
    *status = 1;
    return;
S280:
    switch((int)i99999) {
        case 1: goto S10;
        case 2: goto S20;
        case 3: goto S200;
        default: break;
    }
#undef ftol
}
void zror(status,x,fx,xlo,xhi,qleft,qhi)
int *status;
float *x,*fx,*xlo,*xhi;
unsigned long *qleft,*qhi;
{
    E0000(0,status,x,fx,xlo,xhi,qleft,qhi,
	  fNULL,fNULL,fNULL,fNULL);
}
void stzror(zxlo,zxhi,zabstl,zreltl)
float *zxlo,*zxhi,*zabstl,*zreltl;
{

    E0000(1,fNULL,fNULL,fNULL,fNULL,fNULL,fNULL,fNULL,
	  zabstl,zreltl,zxhi,zxlo);

}

/* 
   float fsign(num, sign)
   float num, sign;
/* Transfers sign of argument sign to argument num */
/* { 
   if ( ( sign>0.0 && num<0.0 ) || ( sign<0.0 && num>0.0 ) )
   return -num;
   else return num;
   }
   */
